## vim:ts=4:et:nowrap
##
##---------------------------------------------------------------------------##
##
## PySol -- a Python Solitaire game
##
## Copyright (C) 2000 Markus Franz Xaver Johannes Oberhumer
## Copyright (C) 1999 Markus Franz Xaver Johannes Oberhumer
## Copyright (C) 1998 Markus Franz Xaver Johannes Oberhumer
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; see the file COPYING.
## If not, write to the Free Software Foundation, Inc.,
## 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
##
## Markus F.X.J. Oberhumer
## <markus.oberhumer@jk.uni-linz.ac.at>
## http://wildsau.idv.uni-linz.ac.at/mfx/pysol.html
##
##---------------------------------------------------------------------------##


# imports
import sys, re, os, string, types

# PySol imports
from mfxutil import Struct, latin1_to_ascii
from resource import CSI


# /***********************************************************************
# // constants
# ************************************************************************/

# GameInfo constants
class GI:
    # game category - these *must* match the cardset CSI.TYPE_xxx
    GC_FRENCH        = CSI.TYPE_FRENCH
    GC_HANAFUDA      = CSI.TYPE_HANAFUDA
    GC_TAROT         = CSI.TYPE_TAROT
    GC_MAHJONGG      = CSI.TYPE_MAHJONGG
    GC_HEXADECK      = CSI.TYPE_HEXADECK

    # game type
    GT_1DECK_TYPE    = 0
    GT_2DECK_TYPE    = 1
    GT_3DECK_TYPE    = 2
    GT_4DECK_TYPE    = 3
    GT_BAKERS_DOZEN  = 4
    GT_BELEAGUERED_CASTLE = 5
    GT_CANFIELD      = 6
    GT_FAN_TYPE      = 7
    GT_FORTY_THIEVES = 8
    GT_FREECELL      = 9
    GT_GOLF          = 10
    GT_GYPSY         = 11
    GT_HANAFUDA      = 12
    GT_HEXADECK      = 13
    GT_KLONDIKE      = 14
    GT_MAHJONGG      = 15
    GT_MONTANA       = 16
    GT_PAIRING_TYPE  = 17
    GT_POKER_TYPE    = 18
    GT_PUZZLE_TYPE   = 19
    GT_ROW_TYPE      = 20
    GT_SIMPLE_TYPE   = 21
    GT_SPIDER        = 22
    GT_TAROT         = 23
    GT_TERRACE       = 24
    GT_YUKON         = 25
    # extra flags
    GT_CHILDREN      = 1 << 12
    GT_OPEN          = 1 << 13
    GT_ORIGINAL      = 1 << 14
    GT_POPULAR       = 1 << 15
    GT_RELAXED       = 1 << 16
    GT_SCORE         = 1 << 17      # game has some type of scoring
    GT_SEPARATE_DECKS = 1 << 18
    GT_HIDDEN        = 1 << 31      # not visible in menus, but games can be loaded

    # game time
    TIME_QUICK  = 0
    TIME_MEDIUM = 1
    TIME_LONG   = 2

    # game difficulty
    ##GD_EASY   = 0
    ##GD_MEDIUM = 1
    ##GD_HARD   = 2

    # skill or chance
    ##SOC_CHANCE_ONLY   = 0
    ##SOC_MOSTLY_CHANCE = 1
    ##SOC_MODERATE      = 2
    ##SOC_MOSTLY_SKILL  = 3
    ##SOC_DIFFICULT     = 4

    # chance of winning
    ##COW_VERY_LOW  = 0
    ##COW_LOW       = 1
    ##COW_MEDIUM    = 2
    ##COW_HIGH      = 3
    ##COW_VERY_HIGH = 4


    SELECT_GAME_BY_TYPE = (
        ("Baker's Dozen type",  lambda gi, gt=GT_BAKERS_DOZEN: gi.si.game_type == gt),
        ("Beleaguered Castle type", lambda gi, gt=GT_BELEAGUERED_CASTLE: gi.si.game_type == gt),
        ("Canfield type",       lambda gi, gt=GT_CANFIELD: gi.si.game_type == gt),
        ("Fan type",            lambda gi, gt=GT_FAN_TYPE: gi.si.game_type == gt),
        ("Forty Thieves type",  lambda gi, gt=GT_FORTY_THIEVES: gi.si.game_type == gt),
        ("FreeCell type",       lambda gi, gt=GT_FREECELL: gi.si.game_type == gt),
        ("Golf type",           lambda gi, gt=GT_GOLF: gi.si.game_type == gt),
        ("Gypsy type",          lambda gi, gt=GT_GYPSY: gi.si.game_type == gt),
        ("Klondike type",       lambda gi, gt=GT_KLONDIKE: gi.si.game_type == gt),
        ("Montana type",        lambda gi, gt=GT_MONTANA: gi.si.game_type == gt),
        ("Pairing type",        lambda gi, gt=GT_PAIRING_TYPE: gi.si.game_type == gt),
        ("Simple games",        lambda gi, gt=GT_SIMPLE_TYPE: gi.si.game_type == gt),
        ("Spider type",         lambda gi, gt=GT_SPIDER: gi.si.game_type == gt),
        ("Terrace type",        lambda gi, gt=GT_TERRACE: gi.si.game_type == gt),
        ("Yukon type",          lambda gi, gt=GT_YUKON: gi.si.game_type == gt),
        ("One-Deck games",      lambda gi, gt=GT_1DECK_TYPE: gi.si.game_type == gt),
        ("Two-Deck games",      lambda gi, gt=GT_2DECK_TYPE: gi.si.game_type == gt),
        ("Three-Deck games",    lambda gi, gt=GT_3DECK_TYPE: gi.si.game_type == gt),
        ("Four-Deck games",     lambda gi, gt=GT_4DECK_TYPE: gi.si.game_type == gt),
    )

    SELECT_SPECIAL_GAME_BY_TYPE = (
        ("Hanafuda type",       lambda gi, gt=GT_HANAFUDA: gi.si.game_type == gt),
        ("Hex A Deck type",     lambda gi, gt=GT_HEXADECK: gi.si.game_type == gt),
        ("Mahjongg type",       lambda gi, gt=GT_MAHJONGG: gi.si.game_type == gt),
        ("Poker type",          lambda gi, gt=GT_POKER_TYPE: gi.si.game_type == gt),
        ("Puzzle type",         lambda gi, gt=GT_PUZZLE_TYPE: gi.si.game_type == gt),
        ("Tarot type",          lambda gi, gt=GT_TAROT: gi.si.game_type == gt),
    )

    # These obsolete gameids have been used in previous versions of
    # PySol and are no longer supported because of internal changes
    # (mainly rule changes). The game has been assigned a new id.
    PROTECTED_GAMES = {
         22:  106,              # Double Canfield
         32:  901,              # La Belle Lucie (Midnight Oil)
         52:  903,              # Aces Up
         68:  904,              # Milligan Harp
         72:  115,              # Little Forty
         75:  126,              # Red and Black
         82:  901,              # La Belle Lucie (Midnight Oil)
         88:  902,              # Trefoil
        121:  905,              # Mississippi
        122:  906,              # Griffon
        155: 5034,              # Mahjongg - Flying Dragon
        156: 5035,              # Mahjongg - Fortress Towers
        262:  105,              # Canfield
    }

    GAMES_BY_COMPATIBILITY = (
        # Atari ST Patience game v2.13 (we have 10 out of 10 games)
        ("Atari ST Patience", (1, 3, 4, 7, 12, 14, 15, 16, 17, 39,)),
        # Gnome AisleRiot 1.0.51 (we have 26 out of 32 games)
        ("Gnome AisleRiot", (
            2, 8, 11, 19, 27, 29, 33, 34, 35, 40, 41, 42, 43, 58,
            59, 92, 93, 94, 95, 96, 100, 105, 111, 112, 113, 903,
        )),
        # KDE KMahjongg 0.4.1 from KDE 1.1.2 (we have 1 out of 4 games)
        ###("KDE KMahjongg", (5019,)),
        # KDE Patience 0.7.3 from KDE 1.1.2 (we have 6 out of 9 games)
        ("KDE Patience", (2, 7, 8, 18, 256, 903,)),
        # xpat2 1.06 (we have 14 out of 16 games)
        ("xpat2", (
            1, 2, 8, 9, 11, 31, 54, 63, 89, 105, 901, 256, 345, 903,
        )),
    )

    GAMES_BY_PYSOL_VERSION = (
        ("1.00", (1, 2, 3, 4)),
        ("1.01", (5, 6)),
        ("1.02", (7, 8, 9)),
        ("1.03", (10, 11, 12, 13)),
        ("1.10", (14,)),
        ("1.11", (15, 16, 17)),
        ("2.00", (256, 257)),
        ("2.01", (258, 259, 260, 261)),
        ("2.02", (105,)),
        ("2.90", (18, 19, 20, 21, 106, 23, 24, 25, 26, 27,
                  28, 29, 30, 31, 901, 33, 34, 35, 36)),
        ("2.99", (37,)),
        ("3.00", (38, 39,
                  40, 41, 42, 43,     45, 46, 47, 48, 49,
                  50, 51,903, 53, 54, 55, 56, 57, 58, 59,
                  60, 61, 62, 63, 64, 65, 66, 67,904, 69,
                  70, 71,115, 73, 74,126, 76, 77, 78, 79,
                  80, 81,     83, 84, 85, 86, 87,902, 89,
                  90, 91, 92, 93, 94, 95, 96, 97, 98, 99,
                  100, 101, 102, 103, 104, 107, 108,)),
        ("3.10", (109, 110, 111, 112, 113, 114, 116, 117, 118, 119,
                  120, 905, 906, 123, 124, 125, 127)),
        ("3.20", (128, 129, 130, 131, 132, 133, 134, 135, 136, 137,
                  138, 139, 140, 141, 142,
                  345, 346, 347, 348, 349, 350, 351, 352)),
        ("3.21", (143, 144)),
        ("3.30", (145, 146, 147, 148, 149, 150, 151)),
        ("3.40", (152, 153, 154)),
        ("4.00", (5034, 5035, 157, 158, 159, 160, 161, 162, 163, 164)),
        ("4.10", range(5001, 5034) + range(5036, 5103)),
    )

    # deprecated - the correct way is to or a GI.GT_XXX flag
    # in the registerGame() call
    _CHILDREN_GAMES = [16, 33, 55, 90, 91, 96, 97, 903,]
    _OPEN_GAMES = [
        5, 6, 8, 9, 26, 31, 45, 46, 50, 53, 63, 64,
        77, 85, 86, 96, 116, 117, 118, 258,
    ]
    _ORIGINAL_GAMES = []
    _POPULAR_GAMES = [
        1, 2, 7, 8, 11, 12, 13, 19, 31, 36, 38, 59, 105, 903, 5034,
    ]
    _RELAXED_GAMES = []


    #
    # sanity check
    #

    def assertGI(self, manager=None):
        for key, games in self.GAMES_BY_COMPATIBILITY:
            print key, len(games), "games"
            for id in games:
                assert type(id) is types.IntType
                if self.PROTECTED_GAMES.get(id):
                    print id; assert 0
                if manager and not manager.get(id):
                    print id; assert 0
        #
        all_games = []
        for key, games in self.GAMES_BY_PYSOL_VERSION:
            print key, len(games), "games"
            for id in games:
                assert type(id) is types.IntType
                if self.PROTECTED_GAMES.get(id):
                    print id; assert 0
                if id in all_games:
                    print id; assert 0
                if manager and not manager.get(id):
                    print id; assert 0
            all_games.extend(list(games))
        print "total:", len(all_games), "games"
        if manager:
            g = manager.getGamesIdSortedById()
            g = filter(lambda id, m=manager: m.get(id).plugin == 0, g)
            assert len(all_games) == len(g)


# /***********************************************************************
# // core games database
# ************************************************************************/

class GameInfoException(Exception):
    pass

class GameInfo(Struct):
    def __init__(self, id, gameclass, name,
                 game_type, decks, redeals,
                 # keyword arguments:
                 si={}, category=0, short_name=None,
                 suits=range(4), ranks=range(13), trumps=()):
        #
        ncards = decks * (len(suits) * len(ranks) + len(trumps))
        game_flags = game_type & ~1023
        game_type = game_type & 1023
        if os.name == "mac":
            name = latin1_to_ascii(name)
        if not short_name:
            short_name = name
        #
        if not (1 <= id <= 32767):
            raise GameInfoException, name + ": invalid game ID " + str(id)
        if not (1 <= decks <= 4):
            raise GameInfoException, name + ": invalid number of decks " + str(id)
        if not name or not (2 <= len(short_name) <= 25):
            raise GameInfoException, name + ": invalid game name"
        if GI.PROTECTED_GAMES.get(id):
            raise GameInfoException, name + ": protected game ID " + str(id)
        #
        for f, l in ((GI.GT_CHILDREN, GI._CHILDREN_GAMES),
                     (GI.GT_OPEN, GI._OPEN_GAMES),
                     (GI.GT_ORIGINAL, GI._ORIGINAL_GAMES),
                     (GI.GT_POPULAR, GI._POPULAR_GAMES),
                     (GI.GT_RELAXED, GI._RELAXED_GAMES)):
            if (game_flags & f) and (id not in l):
                l.append(id)
            elif not (game_flags & f) and (id in l):
                game_flags = game_flags | f
        #
        if not (1 <= category <= 5):
            if game_type == GI.GT_HANAFUDA:
                category = GI.GC_HANAFUDA
            elif game_type == GI.GT_TAROT:
                category = GI.GC_TAROT
            elif game_type == GI.GT_MAHJONGG:
                category = GI.GC_MAHJONGG
            elif game_type == GI.GT_HEXADECK:
                category = GI.GC_HEXADECK
            else:
                category = GI.GC_FRENCH
        # si is the SelectionInfo struct that will be queried by
        # the "select game" dialogs. It can be freely modified.
        gi_si = Struct(game_type=game_type, game_flags=game_flags,
                       decks=decks, redeals=redeals, ncards=ncards)
        gi_si.update(si)
        #
        Struct.__init__(self, id=id, gameclass=gameclass,
                        name=name, short_name=short_name,
                        decks=decks, redeals=redeals, ncards=ncards,
                        category=category,
                        suits=suits, ranks=ranks, trumps=trumps,
                        si=gi_si, rules_filename=None, plugin=0)


class __GameManager:
    def __init__(self):
        self.__selected_key = -1
        self.__games = {}
        self.__games_by_id = None
        self.__games_by_name = None
        self.__games_by_short_name = None
        self.__all_games = {}   # includes hidden games
        self.loading_plugin = 0
        self.registered_game_types = {}

    def getSelected(self):
        return self.__selected_key

    def setSelected(self, gameid):
        assert self.__all_games.has_key(gameid)
        self.__selected_key = gameid

    def get(self, key):
        return self.__all_games.get(key)

    def register(self, gi):
        if not isinstance(gi, GameInfo):
            raise GameInfoException, "wrong GameInfo class"
        gi.plugin = self.loading_plugin
        if self.__all_games.has_key(gi.id):
            raise GameInfoException, "duplicate game ID " + str(gi.id)
#%ifndef BUNDLE
        if 1:
            for id, game in self.__all_games.items():
                if gi.gameclass is game.gameclass:
                    raise GameInfoException, "duplicate game class " + str(gi.id)
                if gi.name == game.name:
                    raise GameInfoException, "duplicate game name " + str(gi.id) + ": " + game.name
#%endif
        self.__all_games[gi.id] = gi
        if not (gi.si.game_flags & GI.GT_HIDDEN):
            self.__games[gi.id] = gi
            # invalidate sorted lists
            self.__games_by_id = None
            self.__games_by_name = None
            # update registry
            k = gi.si.game_type
            self.registered_game_types[k] = self.registered_game_types.get(k, 0) + 1


    #
    # access games database - we do not expose hidden games
    #

    def getGamesIdSortedById(self):
        if self.__games_by_id is None:
            self.__games_by_id = self.__games.keys()
            self.__games_by_id.sort()
        return self.__games_by_id[:]

    def getGamesIdSortedByName(self):
        if self.__games_by_name is None:
            list1a, list1b = [], []
            for id, gi in self.__games.items():
                name = string.lower(latin1_to_ascii(gi.name))
                list1a.append((name, id))
                if gi.name != gi.short_name:
                    name = string.lower(latin1_to_ascii(gi.short_name))
                list1b.append((name, id))
            list1a.sort()
            list1b.sort()
            list2 = []
            for item in list1a:
                list2.append(item[1])
            self.__games_by_name = list2
            list2 = []
            for item in list1b:
                list2.append(item[1])
            self.__games_by_short_name = list2
        return self.__games_by_name[:]

    def getGamesIdSortedByShortName(self):
        if self.__games_by_name is None:
            self.getGamesIdSortedByName()
        return self.__games_by_short_name[:]


# /***********************************************************************
# //
# ************************************************************************/

# the global game database (the single instance of __GameManager)
GAME_DB = __GameManager()


def registerGame(gameinfo):
    GAME_DB.register(gameinfo)
    return gameinfo


def loadGame(modname, filename, plugin=1):
    ##print "load game", modname, filename
    GAME_DB.loading_plugin = plugin
#%ifdef BUNDLE
    #$unbundle#execfile(filename, globals(), globals())
#%else
    if sys.modules.has_key("gamedb"):
        # unbundled
        import imp
        module = imp.load_source(modname, filename)
    else:
        # bundled
        execfile(filename, globals(), globals())
#%endif

