## vim:ts=4:et:nowrap
##
##---------------------------------------------------------------------------##
##
## PySol -- a Python Solitaire game
##
## Copyright (C) 2000 by T. Kirk
## Copyright (C) 2000 Markus Franz Xaver Johannes Oberhumer
## Copyright (C) 1999 Markus Franz Xaver Johannes Oberhumer
## Copyright (C) 1998 Markus Franz Xaver Johannes Oberhumer
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; see the file COPYING.
## If not, write to the Free Software Foundation, Inc.,
## 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
##
## Markus F.X.J. Oberhumer
## <markus.oberhumer@jk.uni-linz.ac.at>
## http://wildsau.idv.uni-linz.ac.at/mfx/pysol.html
##
## T. Kirk
## <grania@mailcity.com>
## http://www.inetarena.com/~grania
##
##---------------------------------------------------------------------------##


# Imports
import sys

# PySol imports
if sys.modules.has_key("pysoltk"):
    from gamedb import registerGame, GameInfo, GI
    from util import *
    from stack import *
    from game import Game
    from layout import Layout
    from hint import AbstractHint, DefaultHint, CautiousDefaultHint
    from pysoltk import MfxCanvasText, getFont


# /***********************************************************************
# // Tarot Foundation Stacks
# ************************************************************************/

class ImperialTrump_Foundation(SS_FoundationStack):
    def acceptsCards(self, from_stack, cards):
        if not SS_FoundationStack.acceptsCards(self, from_stack, cards):
            return 0
        return cards[-1].rank < len(self.game.s.foundations[4].cards)


# /***********************************************************************
# // Tarot Row Stacks
# ************************************************************************/

class Tarot_OpenStack(OpenStack):
    def __init__(self, x, y, game, yoffset, **cap):
        kwdefault(cap, max_move=999999)
        apply(OpenStack.__init__, (self, x, y, game), cap)
        self.CARD_YOFFSET = yoffset


class Skiz_RowStack(SequenceRowStack):
    def _isSequence(self, cards):
        return isRankSequence(cards, self.cap.mod, self.cap.dir)

    def acceptsCards(self, from_stack, cards):
        if not self.basicAcceptsCards(from_stack, cards):
            return 0
        stackcards = self.cards
        if not stackcards:
            if cards[0].suit == len(self.game.gameinfo.suits):
                return cards[0].rank == len(self.game.gameinfo.trumps) - 1
            else:
                return cards[0].rank == len(self.game.gameinfo.ranks) - 1
        return stackcards[-1].suit == cards[0].suit and stackcards[-1].rank - 1 == cards[0].rank


class Pagat_RowStack(SequenceRowStack):
    def _isSequence(self, cards):
        return isRankSequence(cards, self.cap.mod, self.cap.dir)

    def acceptsCards(self, from_stack, cards):
        if not self.basicAcceptsCards(from_stack, cards):
            return 0
        stackcards = self.cards
        if not stackcards:
            return 1
        return stackcards[-1].suit == cards[0].suit and stackcards[-1].rank - 1 == cards[0].rank


class TrumpWild_RowStack(Tarot_OpenStack):
    def acceptsCards(self, from_stack, cards):
        # print self.cap
        if not self.basicAcceptsCards(from_stack, cards):
            return 0
        stackcards = self.cards
        if not stackcards:
            if cards[0].suit == len(self.game.gameinfo.suits):
                return cards[0].rank == len(self.game.gameinfo.trumps) - 1
            else:
                return cards[0].rank == len(self.game.gameinfo.ranks) - 1
        if cards[0].rank != stackcards[-1].rank - 1:
            return 0
        elif cards[0].color == 2 or stackcards[-1].color == 2:
            return 1
        else:
            return cards[0].color != stackcards[-1].color


class TrumpOnly_RowStack(Tarot_OpenStack):
    def acceptsCards(self, from_stack, cards):
        # print self.cap
        if not self.basicAcceptsCards(from_stack, cards):
            return 0
        stackcards = self.cards
        if not stackcards:
            return cards[0].suit == len(self.game.gameinfo.suits)
        return cards[0].color == 2 and cards[0].rank == stackcards[-1].rank - 1

    def getBottomImage(self):
        return self.game.app.images.getReserveBottom()


class Tarot_AC_RowStack(Tarot_OpenStack):
    def acceptsCards(self, from_stack, cards):
        if not self.basicAcceptsCards(from_stack, cards):
            return 0
        stackcards = self.cards
        if not stackcards:
            return 1
        if cards[0].rank != stackcards[-1].rank - 1:
            return 0
        elif cards[0].color == 2 or stackcards[-1].color == 2:
            return 1
        else:
            return cards[0].color != stackcards[-1].color


class Excuse_RowStack(Tarot_OpenStack):
    def acceptsCards(self, from_stack, cards):
        if not self.basicAcceptsCards(from_stack, cards):
            return 0
        stackcards = self.cards
        if not stackcards:
            return 0
        return cards[0].rank == stackcards[-1].rank - 1


class WheelOfFortune_RowStack(Tarot_OpenStack):
    def acceptsCards(self, from_stack, cards):
        if not self.basicAcceptsCards(from_stack, cards):
            return 0
        stackcards = self.cards
        if not stackcards:
            return 1
        return ((cards[0].suit == stackcards[-1].suit)
                and (cards[0].rank == stackcards[-1].rank - 1))

    def getBottomImage(self):
        return self.game.app.images.getReserveBottom()



# /***********************************************************************
# //
# ************************************************************************/

class AbstractTarotGame(Game):
    def getCardFaceImage(self, deck, suit, rank):
        return self.app.images.getFace(deck, suit, rank)


# /***********************************************************************
# // Wheel of Fortune
# ************************************************************************/

class WheelOfFortune(AbstractTarotGame):

    #
    # Game layout
    #

    def createGame(self):
        l, s = Layout(self), self.s
        font = getFont("canvas_card", cardw=l.CW)

        # Set window size
        self.setSize(l.XM + l.XS * 11.5, l.YM + l.YS * 5.5)

        # Create wheel
        xoffset = (1, 2, 3, 3.9, 3, 2, 1, 0, -1, -2, -3,
                    -3.9, -3, -2, -1, 0, -2, -1, 0, 1, 2)
        yoffset = (0.2, 0.5, 1.1, 2.2, 3.3, 3.9, 4.2, 4.4,
                    4.2, 3.9, 3.3, 2.2, 1.1, 0.5, 0.2, 0,
                    1.8, 2.1, 2.2, 2.4, 2.6)
        x = l.XM + l.XS * 4
        y = l.YM
        for i in range(21):
            x0 = x + xoffset[i] * l.XS
            y0 = y + yoffset[i] * l.YS
            s.rows.append(WheelOfFortune_RowStack(x0, y0, self, yoffset=l.CH/4,
                          max_cards=2, max_accept=1))
        self.setRegion(s.rows, (0, 0, l.XS * 9, 999999))

        # Create foundations
        x = self.width - l.XS * 2
        y = l.YM
        s.foundations.append(SS_FoundationStack(x, y, self, 0, max_cards=14))
        x = x + l.XS
        s.foundations.append(SS_FoundationStack(x, y, self, 1, max_cards=14))
        y = y + l.YS
        s.foundations.append(SS_FoundationStack(x, y, self, 3, max_cards=14))
        x = x - l.XS
        s.foundations.append(SS_FoundationStack(x, y, self, 2, max_cards=14))
        x = x + l.XS * 0.5
        y = y + l.YS
        s.foundations.append(SS_FoundationStack(x, y, self, 4, max_cards=22))

        # Create talon
        x = self.width - l.XS
        y = self.height - l.YS * 1.5
        s.talon = WasteTalonStack(x, y, self, num_deal=2, max_rounds=1)
        l.createText(s.talon, "nn")
        x = x - l.XS
        s.waste = WasteStack(x, y, self)
        l.createText(s.waste, "nn")

        # Define stack groups
        l.defaultStackGroups()

    #
    # Game over rides
    #

    def startGame(self):
        assert len(self.s.talon.cards) == 78
        self.startDealSample()
        self.s.talon.dealRow(frames=3)
        self.s.talon.dealCards()

    def shallHighlightMatch(self, stack1, card1, stack2, card2):
        return 0


# /***********************************************************************
# // Imperial Trumps
# ************************************************************************/

class ImperialTrumps(AbstractTarotGame):

    #
    # Game layout
    #

    def createGame(self):
        l, s = Layout(self), self.s
        font = getFont("canvas_card", cardw=l.CW)

        # Set window size
        self.setSize(l.XM + l.XS * 8, l.YM + l.YS * 5)


        # Create foundations
        x = l.XM + l.XS * 3
        y = l.YM
        for i in range(4):
            s.foundations.append(ImperialTrump_Foundation(x, y, self, i, max_cards=14))
            x = x + l.XS
        s.foundations.append(SS_FoundationStack(x, y, self, 4, max_cards=22))

        # Create talon
        x = l.XM
        s.talon = WasteTalonStack(x, y, self, num_deal=1, max_rounds=-1)
        l.createText(s.talon, "ss")
        x = x + l.XS
        s.waste = WasteStack(x, y, self)
        l.createText(s.waste, "ss")

        # Create rows
        x = l.XM
        y = l.YM + int(round(l.YS * 1.25))
        for i in range(8):
            s.rows.append(TrumpWild_RowStack(x, y, self,
                          max_accept=99, yoffset = l.YOFFSET))
            x = x + l.XS
        self.setRegion(s.rows, (0, y, 999999, 999999))

        # Define stack groups
        l.defaultStackGroups()

    #
    # Game over rides
    #

    def startGame(self, reverse=1):
        assert len(self.s.talon.cards) == 78
        for i in range(1, len(self.s.rows)):
            self.s.talon.dealRow(rows=self.s.rows[i:], flip=0, frames=0)
        self.startDealSample()
        self.s.talon.dealRow(reverse=reverse)
        self.s.talon.dealCards()

    def shallHighlightMatch(self, stack1, card1, stack2, card2):
        return 0


# /***********************************************************************
# // Pagat
# ************************************************************************/

class Pagat(AbstractTarotGame):

    #
    # Game layout
    #

    def createGame(self):
        l, s = Layout(self), self.s
        font = getFont("canvas_card", cardw=l.CW)

        # Set window size
        h = max(3 * l.YS, 20 * l.YOFFSET)
        self.setSize(l.XM + 12 * l.XS, l.YM + l.YS + h)

        # Create foundations
        x = l.XM + l.XS * 3.5
        y = l.YM
        s.foundations.append(SS_FoundationStack(x, y, self, 0, max_cards=14))
        x = x + l.XS
        s.foundations.append(SS_FoundationStack(x, y, self, 1, max_cards=14))
        x = x + l.XS
        s.foundations.append(SS_FoundationStack(x, y, self, 4, max_cards=22))
        x = x + l.XS
        s.foundations.append(SS_FoundationStack(x, y, self, 2, max_cards=14))
        x = x + l.XS
        s.foundations.append(SS_FoundationStack(x, y, self, 3, max_cards=14))

        # Create reserves
        x = l.XM
        for i in range(3):
            s.reserves.append(ReserveStack(x, y, self))
            x = x + l.XS
        x = x + l.XS * 6
        for i in range(3):
            s.reserves.append(ReserveStack(x, y, self))
            x = x + l.XS

        # Create rows
        x = l.XM
        y = l.YM + l.YS * 1.1
        for i in range(12):
            s.rows.append(Pagat_RowStack(x, y, self))
            x = x + l.XS
        self.setRegion(s.rows, (0, int(y), 999999, 999999))

        # Create talon
        s.talon = InitialDealTalonStack(l.XM, self.height-l.YS, self)

        # Define stack groups
        l.defaultStackGroups()

    #
    # Game over rides
    #

    def startGame(self):
        assert len(self.s.talon.cards) == 78
        for i in range(6):
            self.s.talon.dealRow(flip=1, frames=0)
        self.startDealSample()
        self.s.talon.dealRow(rows = self.s.rows[3:9])
        self.s.talon.dealCards()

    def shallHighlightMatch(self, stack1, card1, stack2, card2):
        return (card1.suit == card2.suit and
                (card1.rank + 1 == card2.rank
                or card2.rank + 1 == card1.rank))


# /***********************************************************************
# // Skiz
# ************************************************************************/

class Skiz(AbstractTarotGame):

    #
    # Game layout
    #

    def createGame(self):
        l, s = Layout(self), self.s
        font = getFont("canvas_card", cardw=l.CW)

        # Set window size
        h = max(3 * l.YS, 20 * l.YOFFSET)
        self.setSize(l.XM + 12 * l.XS, l.YM + l.YS + h)

        # Create foundations
        x = l.XM + l.XS * 3.5
        y = l.YM
        s.foundations.append(SS_FoundationStack(x, y, self, 0, max_cards=14))
        x = x + l.XS
        s.foundations.append(SS_FoundationStack(x, y, self, 1, max_cards=14))
        x = x + l.XS
        s.foundations.append(SS_FoundationStack(x, y, self, 4, max_cards=22))
        x = x + l.XS
        s.foundations.append(SS_FoundationStack(x, y, self, 2, max_cards=14))
        x = x + l.XS
        s.foundations.append(SS_FoundationStack(x, y, self, 3, max_cards=14))

        # Create reserves
        x = l.XM
        for i in range(3):
            s.reserves.append(ReserveStack(x, y, self))
            x = x + l.XS
        x = x + l.XS * 6
        for i in range(3):
            s.reserves.append(ReserveStack(x, y, self))
            x = x + l.XS

        # Create rows
        x = l.XM
        y = l.YM + l.YS * 1.1
        for i in range(12):
            s.rows.append(Skiz_RowStack(x, y, self))
            x = x + l.XS
        self.setRegion(s.rows, (0, int(y), 999999, 999999))

        # Create talon
        s.talon = InitialDealTalonStack(l.XM, self.height-l.YS, self)

        # Define stack groups
        l.defaultStackGroups()

    #
    # Game over rides
    #

    def startGame(self):
        assert len(self.s.talon.cards) == 78
        for i in range(6):
            self.s.talon.dealRow(flip=1, frames=0)
        self.startDealSample()
        self.s.talon.dealRow(rows = self.s.rows[3:9])
        self.s.talon.dealCards()

    def shallHighlightMatch(self, stack1, card1, stack2, card2):
        return (card1.suit == card2.suit and
                (card1.rank + 1 == card2.rank
                or card2.rank + 1 == card1.rank))


# /***********************************************************************
# // Fifteen Plus
# ************************************************************************/

class FifteenPlus(AbstractTarotGame):

    #
    # Game layout
    #

    def createGame(self):
        l, s = Layout(self), self.s
        font = getFont("canvas_card", cardw=l.CW)

        # Set window size
        h = max(5 * l.YS, 20 * l.YOFFSET)
        self.setSize(l.XM + 9 * l.XS, l.YM + l.YS + h)

        # Create foundations
        x = self.width - l.XS
        y = l.YM
        s.foundations.append(SS_FoundationStack(x, y, self, 4, max_cards=22))
        y = y + l.YS
        for i in range(4):
            s.foundations.append(SS_FoundationStack(x, y, self, i, max_cards=14))
            y = y + l.YS

        # Create rows
        x = l.XM
        y = l.YM
        for j in range(2):
            for i in range(8):
                s.rows.append(Tarot_AC_RowStack(x, y, self, yoffset=l.YOFFSET,
                              max_move=1, max_accept = 1))
                x = x + l.XS
            x = l.XM
            y = y + l.YS * 3
        self.setRegion(s.rows, (0, 0, l.XM + l.XS * 8, 999999))

        # Create talon
        s.talon = InitialDealTalonStack(l.XM, self.height-l.YS, self)

        # Define stack groups
        l.defaultStackGroups()

    #
    # Game over rides
    #

    def startGame(self):
        assert len(self.s.talon.cards) == 78
        for i in range(2):
            self.s.talon.dealRow(flip=0, frames=0)
        for i in range(2):
            self.s.talon.dealRow(rows = self.s.rows[:15], flip=0, frames=0)
        self.startDealSample()
        self.s.talon.dealRow(frames=3)
        self.s.talon.dealCards()

    def shallHighlightMatch(self, stack1, card1, stack2, card2):
        return (card1.suit == card2.suit and
                (card1.rank + 1 == card2.rank
                or card2.rank + 1 == card1.rank))


# /***********************************************************************
# // Excuse
# ************************************************************************/

class Excuse(AbstractTarotGame):

    #
    # Game layout
    #

    def createGame(self):
        l, s = Layout(self), self.s
        font = getFont("canvas_card", cardw=l.CW)

        # Set window size
        h = max(5 * l.YS, 20 * l.YOFFSET)
        self.setSize(l.XM + 9 * l.XS, l.YM + l.YS + h)

        # Create foundations
        x = self.width - l.XS
        y = l.YM
        s.foundations.append(SS_FoundationStack(x, y, self, 4, max_cards=22))
        y = y + l.YS
        for i in range(4):
            s.foundations.append(SS_FoundationStack(x, y, self, i, max_cards=14))
            y = y + l.YS

        # Create rows
        x = l.XM
        y = l.YM
        for j in range(2):
            for i in range(8):
                s.rows.append(Excuse_RowStack(x, y, self, yoffset=l.YOFFSET,
                              max_move=1, max_accept=1))
                x = x + l.XS
            x = l.XM
            y = y + l.YS * 3
        self.setRegion(s.rows, (0, 0, l.XM + l.XS * 8, 999999))

        # Create talon
        s.talon = InitialDealTalonStack(l.XM, self.height-l.YS, self)

        # Define stack groups
        l.defaultStackGroups()

    def _shuffleHook(self, cards):
        i, n = 0, len(self.s.rows) / 2
        kings = []
        for c in cards:
            if c.rank >= 13:
                kings.append(i)
            i = i + 1
        for i in kings:
            j = i % n
            while j < i:
                if cards[j].rank <= 12:
                    swap = cards[j]
                    cards[j] = cards[i]
                    cards[i] = swap
                    break
                j = j + n
        cards.reverse()
        return cards

    #
    # Game over rides
    #

    def startGame(self):
        assert len(self.s.talon.cards) == 78
        for i in range(2):
            self.s.talon.dealRow(flip=1, frames=0)
        for i in range(2):
            self.s.talon.dealRow(rows = self.s.rows[:15], flip=1, frames=0)
        self.startDealSample()
        self.s.talon.dealRow(frames=3)
        self.s.talon.dealCards()

    def shallHighlightMatch(self, stack1, card1, stack2, card2):
        return (card1.rank + 1 == card2.rank
                or card1.rank - 1 == card2.rank)


# /***********************************************************************
# // Grasshopper
# ************************************************************************/

class Grasshopper(AbstractTarotGame):

    #
    # Game layout
    #

    def createGame(self):
        l, s = Layout(self), self.s
        font = getFont("canvas_card", cardw=l.CW)

        # Set window size
        self.setSize(l.XM * 2 + l.XS * 7, l.YM * 2 + l.YS * 5)

        # Create talon
        x = l.XM
        y = l.YM
        s.talon = WasteTalonStack(x, y, self, num_deal = 1, max_rounds = 2)
        l.createText(s.talon, "ss")
        x = x + l.XS
        s.waste = WasteStack(x, y, self)
        l.createText(s.waste, "ss")

        # Create foundations
        x = x + l.XM + l.XS
        for i in range(4):
            s.foundations.append(SS_FoundationStack(x, y, self, i, max_cards=14))
            x = x + l.XS
        s.foundations.append(SS_FoundationStack(x, y, self, 4, max_cards=22))

        # Create reserve
        x = l.XM
        y = l.YM * 3 + l.YS
        s.reserves.append(OpenStack(x, y, self))
        s.reserves[0].CARD_YOFFSET = l.YOFFSET

        # Create rows
        x = x + l.XM + l.XS
        s.rows.append(TrumpOnly_RowStack(x, y, self,
                        max_accept=999999, yoffset=10))
        x = x + l.XS
        for i in range(5):
            s.rows.append(Tarot_AC_RowStack(x, y, self,
                            max_accept=999999, yoffset=l.YOFFSET))
            x = x + l.XS
        self.setRegion(s.rows, (l.XM + l.XS, l.YM * 2 + l.YS, 999999, 999999))

        # Define stack groups
        l.defaultStackGroups()

    #
    # Game over rides
    #

    def startGame(self, reverse=1):
        assert len(self.s.talon.cards) == 78
        self.startDealSample()
        for i in range(14):
            self.s.talon.dealRow(rows=self.s.reserves, flip=0, frames=3, reverse=reverse)
        self.s.reserves[0].flipMove()
        self.s.talon.dealRow(rows = self.s.rows[1:], reverse=reverse)
        self.s.talon.dealCards()

    def fillStack(self, stack):
        r = self.s.reserves[0]
        if not stack.cards and stack in self.s.rows:
            if r.cards:
                r.moveMove(1, stack)
        if r.cards and not r.cards[-1].face_up:
            r.flipMove()

    def shallHighlightMatch(self, stack1, card1, stack2, card2):
        return ((card1.rank + 1 == card2.rank
                or card1.rank - 1 == card2.rank)
                and card1.color != card2.color)


# /***********************************************************************
# // Double Grasshopper
# ************************************************************************/

class DoubleGrasshopper(AbstractTarotGame):

    #
    # Game layout
    #

    def createGame(self):
        l, s = Layout(self), self.s
        font = getFont("canvas_card", cardw=l.CW)

        # Set window size
        self.setSize(l.XM * 2 + l.XS * 12, l.YM * 2 + l.YS * 5)

        # Create talon
        x = l.XM
        y = l.YM
        s.talon = WasteTalonStack(x, y, self, num_deal = 1, max_rounds = 2)
        l.createText(s.talon, "ss")
        x = x + l.XS
        s.waste = WasteStack(x, y, self)
        l.createText(s.waste, "ss")

        # Create foundations
        x = x + l.XM + l.XS
        for j in range(4):
            for i in range(2):
                s.foundations.append(SS_FoundationStack(x, y, self, j, max_cards = 14))
                x = x + l.XS
        for i in range(2):
            s.foundations.append(SS_FoundationStack(x, y, self, 4, max_cards = 22))
            x = x + l.XS

        # Create reserves
        x = l.XM
        y = l.YM * 3 + l.YS
        s.reserves.append(OpenStack(x, y, self))
        s.reserves[0].CARD_YOFFSET = 10

        # Create rows
        x = x + l.XM + l.XS
        for i in range(2):
            s.rows.append(TrumpOnly_RowStack(x, y, self,
                          max_accept=999999, yoffset=10))
            x = x + l.XS
        for i in range(9):
            s.rows.append(Tarot_AC_RowStack(x, y, self,
                          max_accept=999999, yoffset=l.YOFFSET))
            x = x + l.XS
        self.setRegion(s.rows, (l.XM + l.XS, l.YM * 2 + l.YS, 999999, 999999))

        # Define stack groups
        l.defaultStackGroups()

    #
    # Game over rides
    #

    def startGame(self, reverse=1):
        assert len(self.s.talon.cards) == 78 * 2
        self.startDealSample()
        for i in range(28):
            self.s.talon.dealRow(rows=self.s.reserves, flip=0, frames=3, reverse=reverse)
        self.s.reserves[0].flipMove()
        self.s.talon.dealRow(rows = self.s.rows[2:], reverse=reverse)
        self.s.talon.dealCards()

    def fillStack(self, stack):
        r = self.s.reserves[0]
        if not stack.cards and stack in self.s.rows[2:]:
            if r.cards:
                r.moveMove(1, stack)
        if r.cards and not r.cards[-1].face_up:
            r.flipMove()

    def shallHighlightMatch(self, stack1, card1, stack2, card2):
        return ((card1.rank + 1 == card2.rank
                or card1.rank - 1 == card2.rank)
                and card1.color != card2.color)


# register the game
registerGame(GameInfo(157, WheelOfFortune, "Wheel of Fortune",
                      GI.GT_TAROT, 1, 0,
                      suits=range(4), ranks=range(14), trumps=range(22)))
registerGame(GameInfo(158, ImperialTrumps, "Imperial Trumps",
                      GI.GT_TAROT, 1, -1,
                      suits=range(4), ranks=range(14), trumps=range(22)))
registerGame(GameInfo(159, Pagat, "Pagat",
                      GI.GT_TAROT | GI.GT_OPEN, 1, 0,
                      suits=range(4), ranks=range(14), trumps=range(22)))
registerGame(GameInfo(160, Skiz, "Skiz",
                      GI.GT_TAROT | GI.GT_OPEN, 1, 0,
                      suits=range(4), ranks=range(14), trumps=range(22)))
registerGame(GameInfo(161, FifteenPlus, "Fifteen plus",
                      GI.GT_TAROT, 1, 0,
                      suits=range(4), ranks=range(14), trumps=range(22)))
registerGame(GameInfo(162, Excuse, "Excuse",
                      GI.GT_TAROT | GI.GT_OPEN, 1, 0,
                      suits=range(4), ranks=range(14), trumps=range(22)))
registerGame(GameInfo(163, Grasshopper, "Grasshopper",
                      GI.GT_TAROT, 1, 1,
                      suits=range(4), ranks=range(14), trumps=range(22)))
registerGame(GameInfo(164, DoubleGrasshopper, "Double Grasshopper",
                      GI.GT_TAROT, 2, 1,
                      suits=range(4), ranks=range(14), trumps=range(22)))

