#! /usr/bin/env python
##
## vim:ts=4:et:nowrap
##
##---------------------------------------------------------------------------##
##
## PySol -- a Python Solitaire game
##
## Copyright (C) 2000 Markus Franz Xaver Johannes Oberhumer
## Copyright (C) 1999 Markus Franz Xaver Johannes Oberhumer
## Copyright (C) 1998 Markus Franz Xaver Johannes Oberhumer
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; see the file COPYING.
## If not, write to the Free Software Foundation, Inc.,
## 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
##
## Markus F.X.J. Oberhumer
## <markus.oberhumer@jk.uni-linz.ac.at>
## http://wildsau.idv.uni-linz.ac.at/mfx/pysol.html
##
##---------------------------------------------------------------------------##

# imports
import sys, os, re, string, time, types
import traceback

# PySol imports
from mfxutil import destruct, dumpmem, Struct
from mfxutil import pickle, unpickle, UnpicklingError
from mfxutil import merge_dict, getusername, gethomedir, getprefdir, EnvError
from mfxutil import latin1_to_ascii
from resource import CSI, CardsetConfig, Cardset, CardsetManager
from resource import Tile, TileManager
from resource import Sample, SampleManager
from resource import Music, MusicManager
from util import DataLoader, Timer, cyclops
from util import PACKAGE, VERSION
from util import SubsampledImages
from random import LCRandom64
from gamedb import GI, GAME_DB, loadGame
from pysolaudio import pysolsoundserver, AbstractAudioClient, PysolAudioClient

# Toolkit imports
from pysoltk import tkname, tkversion, wm_withdraw
from pysoltk import bind, unbind_destroy
from pysoltk import MfxDialog, MfxExceptionDialog
from pysoltk import TclError, MfxRoot, MfxCanvas, MfxImages
from pysoltk import PysolMenubar
from pysoltk import PysolProgressBar
from pysoltk import PysolToolbar
from pysoltk import PysolStatusbar
from pysoltk import SelectCardsetByTypeDialogWithPreview
from help import helpAbout


# /***********************************************************************
# // init games database
# ************************************************************************/

#%ifndef BUNDLE

# import the games
path = sys.path[:]
sys.path.insert(0, os.path.join(sys.path[0], "games"))
try:
    names = os.listdir(sys.path[0])
except EnvError:
    names = []
names.sort()
for name in names:
    m = re.search(r"^(.+)\.py$", name)
    if m:
##        try:
        exec "import " + m.group(1)
##        except:
##            traceback.print_exc()
sys.path = path
del path, names, name

#%endif /* BUNDLE */


# /***********************************************************************
# // Redirect all pickable classes into the __main__ module.
# ************************************************************************/

#%ifndef BUNDLE

# We need this so that saved games are compatible between the
# bundled and the unbundled versions. This is because of the fact
# that in the bundled version all classes reside in the module `__main__',
# while there are different modules in the unbundled version, and
# this causes problems with pickling as the Un-/Pickler won't find
# the corresponding module/class combination.

t = {}
for m in ("mfxutil", "random", "move"):
    mod = sys.modules[m]
    for k, v in mod.__dict__.items():
        if type(v) is types.ClassType and not t.has_key(k):
            ##print k, v
            t[k] = 1
            # 1) import the class into __main__ so that we can
            #    load bundled games
            exec "from " + m + " import " + k
            # 2) set the module of the class to "__main__" so that
            #    the bundled version can load our unbundled savegames
            v.__module__ = "__main__"

del t, m, mod, k, v

#%endif /* BUNDLE */


# /***********************************************************************
# // These classes also must reside in module __main__ because of
# // pickling compatibility.
# ************************************************************************/

class Options:
    def __init__(self):
        self.player = "unknown"
        self.version = VERSION
        # options menu:
        self.confirm = 1
        self.autofaceup = 1
        self.autodrop = 0
        self.autodeal = 1
        self.quickplay = 0
        self.undo = 1
        self.hint = 1
        self.highlight_piles = 1
        self.highlight_cards = 1
        self.highlight_samerank = 1
        self.cardset = {
            0:              ("Standard", ""),
            GI.GC_FRENCH:   ("Standard", ""),
            GI.GC_HANAFUDA: ("Kintengu", ""),
            GI.GC_MAHJONGG: ("KMahjongg", ""),
            GI.GC_TAROT:    ("Vienna 2K", ""),
            GI.GC_HEXADECK: ("Hex A Deck", ""),
        }
        self.tabletile_name = "Fade_Green.ppm"  # basename
        self.tablecolor = "#008200"
        self.animations = 2                     # default to Timer based
        self.shadow = 0
        self.shade = 0
        self.hint_sleep = 1.5
        self.demo_sleep = 1.5
        self.demo_logo = 1
        self.demo_score = 0
        self.toolbar = 1
        self.statusbar = 1
        self.sound = 1
        self.sound_sample_volume = 128
        self.sound_music_volume = 96
        # additional startup information
        self.last_gameid = 0        # last game played
        self.last_player = None     # last player
        self.last_save_dir = None   # last directory for load/save
        self.game_holded = 0
        # constants
        self.setConstants()

    # not changeable options
    def setConstants(self):
        self.toolbar_dir = "kde-large"
        self.toolbar_relief = 0
        self.dragcursor = 1
        self.magnetic_mouse = 0             # doesn't work yet anyway
        self.magnetic_mouse_time = 2.0      # seconds
        self.raise_card_sleep = 1.0
        self.highlight_piles_sleep = 1.5
        self.highlight_piles_colors = (None, "#ffc000")
        self.highlight_cards_sleep = 1.5
        self.highlight_cards_colors = (None, "#ffc000", None, "#0000ff")
        self.highlight_samerank_sleep = 1.5
        self.highlight_samerank_colors = (None, "#ffc000", None, "#0000ff")
        self.hintarrow_color = "#303030"

    def copy(self):
        opt = Options()
        merge_dict(opt.__dict__, self.__dict__)
        opt.setConstants()
        return opt


class Statistics:
    def __init__(self):
        self.version = VERSION
        # a dictionary of dictionaries of tuples (won/lost),
        # indexed by player and gameid
        self.stats = {}
        self.demo_stats = {}
        # a dictionary of lists of tuples (key: player)
        self.prev_games = {}
        self.all_prev_games = {}
        self.session_games = {}
        # some simple balance scores (key: gameid)
        self.total_balance = {}     # a dictionary of integers
        self.session_balance = {}   # reset per session
        self.gameid_balance = 0     # reset when changing the gameid

    def new(self):
        return Statistics()

    #
    # player statistics
    #

    def resetStats(self, player, gameid):
        if player is None:
            self.__resetDemoStats(gameid)
            return
        assert player and gameid >= 0
        self.resetPrevGames(player, gameid)
        if not self.stats.has_key(player):
            return
        if gameid == 0:
            del self.stats[player]
        else:
            w0, l0 = self.getStats(player, 0)
            w1, l1 = self.getStats(player, gameid)
            self.stats[player][0] = (w0 - w1, l0 - l1)
            self.stats[player][gameid] = (0, 0)

    def resetPrevGames(self, player, gameid):
        assert player and gameid >= 0
        if not self.prev_games.has_key(player):
            return
        if gameid == 0:
            del self.prev_games[player]
        else:
            self.prev_games[player] = filter(lambda a, b=gameid: a[0] != b, self.prev_games[player])

    def getStats(self, player, gameid):
        if player is None:
            return self.__getDemoStats(gameid)
        assert player and gameid >= 0
        d = self.stats.get(player)
        if d is None:
            # create an entry for this player
            d = self.stats[player] = {}
        return d.get(gameid, (0,0))

    def updateStats(self, player, game, status):
        won, lost = status > 0, status == 0
        if player is None:
            self.__updateDemoStats(game.id, won, lost)
            return
        assert player and game.id > 0
        w, l = self.getStats(player, 0)               # all games
        self.stats[player][0] = (w + won, l + lost)
        w, l = self.getStats(player, game.id)
        self.stats[player][game.id] = (w + won, l + lost)
        self.updateLog(player, game, status)

    def updateLog(self, player, game, status):
        log = (game.id, game.getFullId(format=0), status,
               game.gstats.start_time, game.gstats.total_elapsed_time,
               VERSION, game.getGameScore())
        if status >= 0:
            if not self.prev_games.has_key(player):
                self.prev_games[player] = []
            self.prev_games[player].append(log)
            if not self.all_prev_games.has_key(player):
                self.all_prev_games[player] = []
            self.all_prev_games[player].append(log)
        if not self.session_games.has_key(player):
            self.session_games[player] = []
        self.session_games[player].append(log)

    #
    # demo statistics (deprecated interface)
    #

    def __resetDemoStats(self, gameid):
        # a dictionary of tuples (won/lost), indexed by gameid
        if gameid == 0:
            self.demo_stats = {}
        else:
            won, lost = self.__getDemoStats(gameid)
            self.__updateDemoStats(gameid, -won, -lost)

    def __getDemoStats(self, gameid):
        assert gameid >= 0
        return self.demo_stats.get(gameid, (0,0))

    def __updateDemoStats(self, gameid, won, lost):
        assert gameid > 0
        w, l = self.__getDemoStats(0)               # all games
        self.demo_stats[0] = (w + won, l + lost)
        w, l = self.__getDemoStats(gameid)
        self.demo_stats[gameid] = (w + won, l + lost)


# /***********************************************************************
# // Application
# // This is the glue between the toplevel window and a Game.
# // Also handles all global resources.
# ************************************************************************/

class Application:
    def __init__(self):
        self.starttimer = Timer("Application.__init__")
        self.gdb = GAME_DB
        self.opt = Options()
        self.stats = Statistics()
        self.splashscreen = 1
        self.debug = 0
        # visual components
        self.top = None                 # the root toplevel window
        self.top_cursor = None          # default cursor
        self.menubar = None
        self.toolbar = None
        self.canvas = None
        self.statusbar = None
        #
        self.game = None
        self.dataloader = None
        self.audio = None
        self.images = None
        self.subsampled_images = None
        self.jokers = []
        self.redeal_images = []
        self.demo_images = []           # demo logos
        self.progress_bg = None
        self.progress_images = []
        self.cardset_manager = CardsetManager()
        self.cardset = None             # current cardset
        self.tabletile_manager = TileManager()
        self.tabletile_index = 0        # current table tile
        self.sample_manager = SampleManager()
        self.music_manager = MusicManager()
        self.music_playlist = []
        self.intro = Struct(
            progress = None,            # progress bar
        )
        # directory names
        home = os.path.normpath(gethomedir())
        config = os.path.normpath(getprefdir(PACKAGE, home))
        if 0 and os.name == "nt":
            # use old location of config dir if it exists
            d = os.path.join(home, ".pysol")
            if d and os.path.isdir(d):
                config = d
        self.dn = Struct(
            home = home,
            config = config,
            plugins = os.path.join(config, "plugins"),
            savegames = os.path.join(config, "savegames"),
            maint = os.path.join(config, "maint"),          # debug
        )
        for k, v in self.dn.__dict__.items():
##            if os.name == "nt":
##                v = os.path.normcase(v)
            v = os.path.normpath(v)
            self.dn.__dict__[k] = v
        # file names
        self.fn = Struct(
            opt   = os.path.join(self.dn.config, "options.dat"),
            stats = os.path.join(self.dn.config, "statistics.dat"),
            holdgame = os.path.join(self.dn.config, "holdgame.dat"),
        )
        for k, v in self.dn.__dict__.items():
            if os.name == "nt":
                v = os.path.normcase(v)
            v = os.path.normpath(v)
            self.fn.__dict__[k] = v
        # random generators
        self.gamerandom = LCRandom64()
        self.miscrandom = LCRandom64()
        # player
        player = getusername()
        if not player:
            player = "unknown"
        player = player[:30]
        self.opt.player = player
        # misc
        self.nextgame = Struct(
            id = 0,                     # start this game
            random = None,              # use this random generator
            loadedgame = None,          # data for loaded game
            startdemo = 0,              # start demo ?
            cardset = None,             # use this cardset
            holdgame = 0,               # hold this game on exit ?
            gstats = None,              # use these gstats (restart)
        )
        self.commandline = Struct(
            loadgame = None,            # load a game ?
        )
        self.demo_counter = 0


    # the PySol mainloop
    def mainloop(self):
        # try to load statistics
        try: self.loadStatistics()
        except: pass
        # startup information
        if self.getGameClass(self.opt.last_gameid):
            self.nextgame.id = self.opt.last_gameid
        # load a holded or saved game
        tmpgame = self.constructGame(1)
        if self.opt.game_holded > 0 and not self.nextgame.loadedgame:
            game = None
            try:
                game = tmpgame._loadGame(self.fn.holdgame, self)
            except:
                if 1 or self.debug: traceback.print_exc()           #bundle#
                game = None
            if game:
                if game.id == self.opt.game_holded and game.gstats.holded:
                    game.gstats.loaded = game.gstats.loaded - 1
                    game.gstats.holded = 0
                    self.nextgame.loadedgame = game
                else:
                    # not a holded game
                    game.destruct()
                    destruct(game)
            game = None
        if self.commandline.loadgame and not self.nextgame.loadedgame:
            try:
                self.nextgame.loadedgame = tmpgame._loadGame(self.commandline.loadgame, self)
                self.nextgame.loadedgame.gstats.holded = 0
            except:
                self.nextgame.loadedgame = None
        self.opt.game_holded = 0
        tmpgame.destruct()
        destruct(tmpgame)
        tmpgame = None
        # create the menubar
        self.menubar = PysolMenubar(self, self.top)
        # create the statusbar (before the toolbar)
        self.statusbar = PysolStatusbar(self.top)
        self.statusbar.show(self.opt.statusbar)
        # create the toolbar
        dir = os.path.join(self.dataloader.dir, "toolbar", self.opt.toolbar_dir)
        self.toolbar = PysolToolbar(self.top, dir=dir)
        self.toolbar.setRelief(self.opt.toolbar_relief)
        self.toolbar.show(self.opt.toolbar)
        if self.intro.progress: self.intro.progress.update(step=1)
        # create the canvas
        self.canvas = MfxCanvas(self.top, bg=self.opt.tablecolor, highlightthickness=0)
        self.canvas.pack(fill="both", expand=1)
        tile = self.tabletile_manager.get(self.tabletile_index)
        self.canvas.setTile(tile.filename)
        self.canvas.setTextColor(tile.text_color)
        try:
            # this is the mainloop
            while 1:
                while self.cardset is None:
                    self.requestCompatibleCardset(self.nextgame.id)
                id, random = self.nextgame.id, self.nextgame.random
                self.nextgame.id, self.nextgame.random = 0, None
                self.runGame(id, random)
                if self.nextgame.holdgame:
                    assert self.nextgame.id <= 0
                    try:
                        self.game.gstats.holded = 1
                        self.game._saveGame(self.fn.holdgame)
                        self.opt.game_holded = self.game.id
                    except: pass
                self.freeGame()
                self.dumpMem("after freeGame")                      #bundle#
                if self.nextgame.id <= 0:
                    break
                # load new cardset
                if self.nextgame.cardset is not self.cardset:
                    self.loadCardset(self.nextgame.cardset, id=self.nextgame.id)
                else:
                    self.requestCompatibleCardset(self.nextgame.id)
        finally:
            # save options
            self.opt.last_gameid = id
            try: self.saveOptions()
            except: pass
            # save statistics
            try: self.saveStatistics()
            except: pass
            # shut down audio
            try: self.audio.destroy()
            except: pass


    def runGame(self, id, random=None):
        self.top.connectApp(self)
        # create game instance
        g = self.getGameClass(id)
        if g is None:
            id = 2          # start Klondike as default game
            random = None
            g = self.getGameClass(id)
            if g is None:
                # start first available game
                id = self.gdb.getGamesIdSortedByName()[0]
                g = self.getGameClass(id)
        gi = self.getGameInfo(id)
        assert g and type(g) is types.ClassType and id > 0
        assert gi is not None and gi.id == id
        self.game = self.constructGame(id)
        self.gdb.setSelected(id)
        self.game.busy = 1
        # create stacks
        self.game.create(self)
        # delete intro progress bar
        if self.intro.progress:
            self.intro.progress.destroy()
            destruct(self.intro.progress)
            self.intro.progress = None
        # connect with game
        self.menubar.connectGame(self.game)
        self.toolbar.connectGame(self.game, self.menubar)
        # start game
        if self.nextgame.loadedgame is not None:
            self.stats.gameid_balance = 0
            self.game.restoreGame(self.nextgame.loadedgame)
            destruct(self.nextgame.loadedgame)
            autoplay = 0
        elif self.nextgame.gstats is not None:
            self.game.gstats = self.nextgame.gstats
            self.game.newGame(random=random, restart=1, autoplay=0)
            autoplay = 1
        else:
            self.stats.gameid_balance = 0
            self.game.newGame(random=random, autoplay=0)
            autoplay = 1
        self.nextgame.loadedgame = None
        self.nextgame.gstats = None
        self.intro.progress = None
        if self.splashscreen and not self.debug:
            status = helpAbout(self, timeout=20000)
            if status == 2:                 # timeout - start a demo
                self.nextgame.startdemo = 1
        self.splashscreen = 0
        if self.nextgame.startdemo:
            self.nextgame.startdemo = 0
            self.game.startDemo()
            self.game.createDemoInfoText()
        elif autoplay:
            self.game.autoPlay()
            self.game.changed_index = self.game.moves.index
            self.game.stats.player_moves = 0
        # enter the Tk mainloop
        self.game.busy = 0
        self.top.mainloop()


    # free game
    def freeGame(self):
        # disconnect from game
        self.toolbar.connectGame(None, None)
        self.menubar.connectGame(None)
        # clean up the canvas
        unbind_destroy(self.canvas)
        self.canvas.deleteAllItems()
        self.canvas.update_idletasks()
        # destruct the game
        if self.game:
            self.game.destruct()
            destruct(self.game)
        self.game = None
        self.top.connectApp(None)


    # debug
    def dumpMem(self, info=""):
#%ifndef BUNDLE
        return
        if info: print "App.dumpMem: " + info
        print "freeGame:\n"; dumpmem(); print
        if 1 and cyclops: cyclops.clear()
        if 1 and cyclops: cyclops.register(self)
        if 1 and cyclops: cyclops.find_cycles()
        if 1 and cyclops: cyclops.show_stats()
        #if 1 and cyclops: cyclops.show_arcs()
        #if 1 and cyclops: cyclops.show_sccs()
        #if 1 and cyclops: cyclops.show_cycles()
        if 1 and cyclops: cyclops.clear()
#%endif
        pass


    #
    # cardset
    #

    def updateCardset(self, id=0):
        cs = self.images.cs
        self.cardset = cs
        self.nextgame.cardset = cs
        # update settings
        self.cardset_manager.setSelected(cs.index)
        # update options
        self.opt.cardset[0] = (cs.name, cs.backname)
        self.opt.cardset[cs.si.type] = (cs.name, cs.backname)
        if id > 0:
            gi = self.getGameInfo(id)
            if gi and not self.checkCompatibleCardset(id):
                self.opt.cardset[gi.category] = (cs.name, cs.backname)
        ###print self.opt.cardset

    def loadCardset(self, cs, id=0, progress=None):
        if cs is None or cs.error or cs is self.cardset:
            return
        if progress is None:
            wm_withdraw(self.top)
            self.top.busyUpdate()
            title = "Loading cardset " + cs.name + "..."
            color = self.opt.tablecolor
            if self.tabletile_index > 0:
                color = "#008200"
            progress = PysolProgressBar(self, self.top, title=title, color=color, bg=self.progress_bg, images=self.progress_images)
        images = MfxImages(self.dataloader, cs)
        try:
            images.load(progress=progress)
            simages = SubsampledImages(images)
            # update
            self.images.destruct()
            destruct(self.images)
            self.images = images
            self.subsampled_images = simages
            self.updateCardset(id)
        except (Exception, TclError, UnpicklingError), ex:
            if 1 or self.debug: traceback.print_exc()               #bundle#
            # restore settings
            self.nextgame.cardset = self.cardset
            if self.cardset:
                self.cardset_manager.setSelected(self.cardset.index)
            images.destruct()
            destruct(images)
            d = MfxExceptionDialog(self.top, ex, title="Cardset load error",
                                   text="Error while loading cardset")
        self.intro.progress = progress
        if self.menubar is not None:
            self.menubar.updateBackgroundImages()

    def checkCompatibleCardset(self, id):
        gi = self.getGameInfo(id)
        if gi is None:
            return None
        gc = gi.category
        cs_type = -1
        if self.cardset:
            cs_type = self.cardset.type
        t = None
        if gc == GI.GC_FRENCH:
            if cs_type not in (CSI.TYPE_FRENCH, CSI.TYPE_TAROT, CSI.TYPE_HEXADECK):
                t = "French"
        elif gc == GI.GC_HANAFUDA:
            if cs_type not in (CSI.TYPE_HANAFUDA,):
                t = "Hanafuda"
        elif gc == GI.GC_TAROT:
            if cs_type not in (CSI.TYPE_TAROT,):
                t = "Tarot"
        elif gc == GI.GC_MAHJONGG:
            if cs_type not in (CSI.TYPE_MAHJONGG,):
                t = "Mahjongg"
        elif gc == GI.GC_HEXADECK:
            if cs_type not in (CSI.TYPE_HEXADECK,):
                t = "Tarot"
        else:
            # we should not come here
            t = "Unknown"
        return t

    def requestCompatibleCardset(self, id):
        t = self.checkCompatibleCardset(id)
        if not t:
            return
        #
        if self.intro.progress:
            self.intro.progress.destroy()
            destruct(self.intro.progress)
            self.intro.progress = None
        #
        gi = self.getGameInfo(id)
        n, bn = self.opt.cardset.get(gi.category, (None, None))
        if n:
            cs = self.cardset_manager.getByName(n)
            if cs:
                cs.updateCardback(backname=bn)
                self.loadCardset(cs, id=id)
                return
        #
        cs = self.cardset
        d = MfxDialog(self.top, title="Incompatible Cardset", bitmap="warning",
                      strings=("OK",), default=0,
                      text="The currently selected cardset\n" + cs.name
                        + "\n\nis not compatible with the game\n"
                        + gi.name + "\n\nPlease select a " + t + " type cardset.")
        #
        strings, default = (None, "OK", "Cancel"), 1
        key = self.cardset.index
        d = SelectCardsetByTypeDialogWithPreview(self.top, title="Please select a " + t + " type cardset",
                app=self, manager=self.cardset_manager, key=key,
                strings=strings, default=default)
        if d.status == 0 and d.button in (0, 1) and d.key >= 0:
            cs = self.cardset_manager.get(d.key)
            self.loadCardset(cs, id=id)


    #
    # options & statistics
    #

    def loadOptions(self):
        opt = unpickle(self.fn.opt)
        if opt:
            merge_dict(self.opt.__dict__, opt.__dict__)
            self.opt.setConstants()

    def saveOptions(self):
        self.opt.version = VERSION
        pickle(self.opt, self.fn.opt)

    def loadStatistics(self):
        stats = unpickle(self.fn.stats)
        if stats:
            merge_dict(self.stats.__dict__, stats.__dict__)
        # start a new session
        self.stats.session_games = {}
        self.stats.session_balance = {}
        self.stats.gameid_balance = 0

    def saveStatistics(self):
        self.stats.version = VERSION
        pickle(self.stats, self.fn.stats, binmode=1)


    #
    # access games database
    #

    def constructGame(self, id):
        gi = self.gdb.get(id)
        if gi is None:
            raise Exception, "Unknown game (id %d)" % id
        return gi.gameclass(gi)

    def getGamesIdSortedById(self):
        return self.gdb.getGamesIdSortedById()

    def getGamesIdSortedByName(self):
        return self.gdb.getGamesIdSortedByName()

    def getGameInfo(self, id):
        return self.gdb.get(id)

    def getGameClass(self, id):
        gi = self.gdb.get(id)
        if gi is None: return None
        return gi.gameclass

    def getGameTitleName(self, id):
        gi = self.gdb.get(id)
        if gi is None: return None
        return gi.name

    def getGameMenuitemName(self, id):
        gi = self.gdb.get(id)
        if gi is None: return None
        return gi.short_name

    def getGameRulesFilename(self, id):
        gi = self.gdb.get(id)
        if gi is None: return None
        if gi.rules_filename is not None:
            return gi.rules_filename
        n = gi.name
        n = re.sub(r"[\[\(].*$", "", n)
        n = latin1_to_ascii(n)
        n = re.sub(r"[^\w]", "", n)
        n = string.lower(n) + ".html"
        gi.rules_filename = n    # cache the filename for next use
        return n

    def getGameSaveName(self, id):
        n = self.getGameTitleName(id)
        if not n: return None
        m = re.search(r"^(.*)([\[\(](\w+).*[\]\)])\s*$", n)
        if m:
            n = m.group(1) + "_" + string.lower(m.group(2))
        n = latin1_to_ascii(n)
        return re.sub(r"[^\w-]", "", n)

    def getRandomGameId(self):
        return self.miscrandom.choice(self.gdb.getGamesIdSortedById())


    #
    # plugins
    #

    def loadPlugins(self, dir):
        if not dir or not os.path.isdir(dir):
            return
        names = os.listdir(dir)
        names.sort()
        for name in names:
            m = re.search(r"^(.+)\.py$", name)
            n = os.path.join(dir, name)
            if m and os.path.isfile(n):
                p = sys.path[:]
                try:
                    loadGame(m.group(1), n)
                except Exception, ex:
                    print "Error loading plugin " + n + ": " + str(ex)
                    sys.stdout.flush()
                    if 1 or self.debug: traceback.print_exc()       #bundle#
                except:
                    if 1 or self.debug: traceback.print_exc()       #bundle#
                    pass
                sys.path = p


    #
    # cardsets
    #

    # read & parse a cardset config.txt file - see class Cardset in util.py
    def _readCardsetConfig(self, dir, filename):
        f = None
        try:
            f = open(filename, "r")
            lines = f.readlines()
        finally:
            if f: f.close()
        lines = map(string.strip, lines)
        if string.index(lines[0], "PySol") != 0:
            return None
        config = CardsetConfig()
        if not self._parseCardsetConfig(config, lines):
            return None
        if config.CARDD > self.top.winfo_screendepth():
            return None
        cs = Cardset()
        cs.dir = dir
        cs.update(config.__dict__)
        return cs

    def _parseCardsetConfig(self, cs, line):
        if len(line) < 6:
            return 0
        # line[0]: magic identifier, possible version information
        fields = filter(None, re.split(r";", line[0]))
        fields = map(string.strip, fields)
        if len(fields) >= 2:
            m = re.search(r"^(\d+)$", fields[1])
            if m: cs.version = int(m.group(1))
        if cs.version >= 3:
            if len(fields) < 5:
                return 0
            cs.ext = fields[2]
            m = re.search(r"^(\d+)$", fields[3])
            if not m: return 0
            cs.type = int(m.group(1))
            m = re.search(r"^(\d+)$", fields[4])
            if not m: return 0
            cs.ncards = int(m.group(1))
        if cs.version >= 4:
            if len(fields) < 6:
                return 0
            styles = string.splitfields(fields[5], ",")
            for s in styles:
                m = re.search(r"^\s*(\d+)\s*$", s)
                if not m: return 0
                s = int(m.group(1))
                if not s in cs.styles:
                    cs.styles.append(s)
        if cs.version >= 5:
            if len(fields) < 7:
                return 0
            m = re.search(r"^(\d+)$", fields[6])
            if not m: return 0
            cs.year = int(m.group(1))
        if len(cs.ext) < 2 or cs.ext[0] != ".":
            return 0
        # line[1]: identifier/name
        if not line[1]: return 0
        cs.ident = line[1]
        m = re.search(r"^(.*;)?([^;]+)$", cs.ident)
        if not m: return 0
        cs.name = string.strip(m.group(2))
        # line[2]: CARDW, CARDH, CARDD
        m = re.search(r"^(\d+)\s+(\d+)\s+(\d+)", line[2])
        if not m: return 0
        cs.CARDW, cs.CARDH, cs.CARDD = int(m.group(1)), int(m.group(2)), int(m.group(3))
        # line[3]: CARD_UP_YOFFSET, CARD_DOWN_YOFFSET, SHADOW_XOFFSET, SHADOW_YOFFSET
        m = re.search(r"^(\d+)\s+(\d+)\s+(\d+)\s+(\d+)", line[3])
        if not m: return 0
        cs.CARD_UP_YOFFSET = int(m.group(1))
        ###cs.CARD_DOWN_YOFFSET = int(m.group(2))   # not used
        cs.SHADOW_XOFFSET = int(m.group(3))
        cs.SHADOW_YOFFSET = int(m.group(4))
        # line[4]: default background
        back = line[4]
        if not back: return 0
        # line[5]: all available backgrounds
        cs.backnames = re.split(r";", line[5])
        cs.backnames = map(string.strip, cs.backnames)
        cs.backnames = filter(None, cs.backnames)
        if back in cs.backnames:
            cs.backindex = cs.backnames.index(back)
        else:
            cs.backnames.insert(0, back)
            cs.backindex = 0
        ##if cs.type != 1: print cs.type, cs.name
        return 1

    def initCardsets(self):
        manager = self.cardset_manager
        # find all available cardsets
        dirs = manager.getSearchDirs(self, ("cardsets", ""), "PYSOL_CARDSETS")
        if self.debug:
            dirs = dirs + manager.getSearchDirs(self, "cardsets*")
        try:
            dirs = dirs + manager.getRegistryDirs(self, "Cardsets")
        except:
            traceback.print_exc()                                   #bundle#
            pass
        ##print dirs
        found, t = [], {}
        for dir in dirs:
            dir = string.strip(dir)
            try:
                names = []
                if dir and os.path.isdir(dir) and not t.has_key(dir):
                    t[dir] = 1
                    names = os.listdir(dir)
                    names.sort()
                for name in names:
                    m = re.search(r"^cardset-", name, re.I)
                    if not m: continue
                    d = os.path.join(dir, name)
                    if not os.path.isdir(d): continue
                    f1 = os.path.join(d, "config.txt")
                    f2 = os.path.join(d, "COPYRIGHT")
                    if os.path.isfile(f1) and os.path.isfile(f2):
                        try:
                            cs = self._readCardsetConfig(d, f1)
                            if cs:
                                ##print cs.name, cs.__dict__
                                back = cs.backnames[cs.backindex]
                                f1 = os.path.join(d, back)
                                f2 = os.path.join(d, "shade" + cs.ext)
                                if os.path.isfile(f1) and os.path.isfile(f2):
                                    found.append(cs)
                        except:
                            if self.debug: traceback.print_exc()    #bundle#
                            pass
            except EnvError, ex:
                pass
        # register cardsets
        for obj in found:
            if not manager.getByName(obj.name):
                manager.register(obj)
                ##print obj.index, obj.name


    #
    # tiles
    #

    def initTiles(self):
        manager = self.tabletile_manager
        # find all available tiles
        dirs = manager.getSearchDirs(self, "tiles*", "PYSOL_TILES")
        try:
            dirs = dirs + manager.getRegistryDirs(self, "Tiles")
        except:
            traceback.print_exc()                                   #bundle#
            pass
        ##print dirs
        ext_re        = re.compile(r"\.((gif)|(ppm))$", re.I)
        text_color_re = re.compile(r"^(.+)-([0-9A-Fa-f]{6})$")
        found, t = [], {}
        for dir in dirs:
            dir = string.strip(dir)
            try:
                names = []
                if dir and os.path.isdir(dir):
                    names = os.listdir(dir)
                    names.sort()
                for name in names:
                    if not name or not ext_re.search(name):
                        continue
                    f = os.path.join(dir, name)
                    if not os.path.isfile(f):
                        continue
                    tile = Tile()
                    tile.filename = f
                    n = ext_re.sub("", string.strip(name))
                    m = text_color_re.search(n)
                    if m:
                        n = m.group(1)
                        tile.text_color = "#" + string.lower(m.group(2))
                    n = re.sub("_", " ", n)
                    tile.name = n
                    key = string.lower(n)
                    if not t.has_key(key):
                        t[key] = 1
                        found.append((n, tile))
            except EnvError, ex:
                pass
        # register tiles
        found.sort()
        for f in found:
            obj = f[1]
            if not manager.getByName(obj.name):
                manager.register(obj)

    #
    # samples / music
    #

    def initResource(self, manager, dirs, ext_re, ResourceClass):
        found, t = [], {}
        for dir in dirs:
            dir = string.strip(dir)
            if dir:
                dir = os.path.normpath(dir)
            try:
                names = []
                if dir and os.path.isdir(dir):
                    names = os.listdir(dir)
                    names.sort()
                for name in names:
                    if not name or not ext_re.search(name):
                        continue
                    f = os.path.join(dir, name)
                    f = os.path.normpath(f)
                    if not os.path.isfile(f):
                        continue
                    obj = ResourceClass()
                    obj.filename = f
                    n = ext_re.sub("", string.strip(name))
                    obj.name = n
                    key = string.lower(n)
                    if not t.has_key(key):
                        t[key] = 1
                        found.append((n, obj))
            except EnvError, ex:
                pass
        # register songs
        found.sort()
        for f in found:
            obj = f[1]
            if not manager.getByName(obj.name):
                manager.register(obj)


    def initSamples(self):
        manager = self.sample_manager
        # find all available samples
        dirs = manager.getSearchDirs(self, ("sound", os.path.join("sound", "extra")))
        ##print dirs
        ext_re = re.compile(r"\.((wav))$", re.I)
        self.initResource(manager, dirs, ext_re, Music)


    def initMusic(self):
        manager = self.music_manager
        # find all available music songs
        dirs = manager.getSearchDirs(self, "music*", "PYSOL_MUSIC")
        try:
            dirs = dirs + manager.getRegistryDirs(self, "Music")
        except:
            traceback.print_exc()                                   #bundle#
            pass
        ##print dirs
        ext_re = re.compile(r"\.((it)|(mod)|(mp3)|(s3m)|(xm))$", re.I)
        self.initResource(manager, dirs, ext_re, Music)


# /***********************************************************************
# // PySol main entry
# ************************************************************************/

def pysol_main(args):
    # create the application
    app = Application()

    # try to create the config directory
    try: os.makedirs(app.dn.config, 0750)
    except: pass
    try: os.mkdir(os.path.join(app.dn.savegames), 0750)
    except: pass
    try: os.mkdir(os.path.join(app.dn.config, "screenshots"), 0750)
    except: pass
    try: os.mkdir(os.path.join(app.dn.config, "tiles"), 0750)
    except: pass

    # init DataLoader
    f = os.path.join("images", "tree", "python.gif")
    app.dataloader = DataLoader(args[0], f)

    # try to load plugins
    if not "--noplugins" in args[1:]:
        for dir in (os.path.join(app.dataloader.dir, "games"),
                    os.path.join(app.dataloader.dir, "plugins"),
                    app.dn.plugins):
            try:
                app.loadPlugins(dir)
            except:
                if app.debug: traceback.print_exc()                 #bundle#
                pass

    # init commandline options (undocumented)
    wm_command = ""
    opt_nosound = 0
    prog = sys.executable
    argv0 = os.path.normpath(args[0])
    if prog and os.path.isfile(prog):
        prog = os.path.abspath(prog)
        if os.path.isfile(argv0):
            wm_command = prog + " " + os.path.abspath(argv0)
    for a in args[1:]:
        if os.path.isfile(a):
            app.commandline.loadgame = a
        elif a == "--nodebug":
            app.debug = 0
        elif a == "--debug":
            app.debug = app.debug + 1
        elif a[:9] == "--bindir=":
            wm_command = wm_command + " " + a
        elif a == "--nosound":
            opt_nosound = 1
            wm_command = wm_command + " " + a
        # ignore other commandline options

    # try to load options
    try: app.loadOptions()
    except EnvError, ex: pass
    except:
        if app.debug: traceback.print_exc()                         #bundle#
        pass

    # print some debug info
    if 1 and not app.debug and app.opt.player == "mfx":             #bundle#
        app.debug = 1                                               #bundle#
    if not app.debug:
        app.dn.maint = None
    if 1 and app.debug >= 2:
        print "prog:", prog
        print "argv0:", argv0
        print "wm_command:", wm_command
        print "sys.executable:", sys.executable
        print "sys.version:", sys.version
        print "sys.platform:", sys.platform
        print "sys.argv:", sys.argv
        print "sys.path:", sys.path
        print "dataloader:", app.dataloader.__dict__
        print "tkname:", tkname
        print "tkversion:", tkversion
    if 1 and app.debug >= 2:
        GI().assertGI(app.gdb)

#%ifndef BUNDLE
    # import Cyclops
    if 1 and app.debug:
        global cyclops
        try:
            import Cyclops
            cyclops = Cyclops.CycleFinder()
            print "Imported Cyclops."
        except ImportError:
            pass
#%endif

    # init audio 1)
    if opt_nosound:
        app.audio = AbstractAudioClient()
    else:
        app.audio = PysolAudioClient()
    app.audio.startServer()

    # init toolkit
    top = MfxRoot(className=PACKAGE)
    top.wm_group(top)
    top.wm_title(PACKAGE + " " + VERSION)
    top.wm_iconname(PACKAGE + " " + VERSION)
    top.wm_minsize(200, 200)
    top.wm_protocol("WM_DELETE_WINDOW", top.wmDeleteWindow)
    if wm_command:
        top.wm_command(wm_command)
    sw, sh, sd = top.winfo_screenwidth(), top.winfo_screenheight(), top.winfo_screendepth()
    if sw < 640 or sh < 480 or sd < 8:
        wm_withdraw(top)
        top.update()
        d = MfxDialog(top, title=PACKAGE + " init error",
                      text=PACKAGE + " requires at least\n640x480 in 256 colors.",
                      bitmap="error", strings=("Quit",))
        return 1
    if 1:
        # set expected window size to assist the layout of the window manager
        top.config(width=min(800,sw-64), height=min(600,sh-64), bg=app.opt.tablecolor)
    else:
        top.config(bg=app.opt.tablecolor)
    try:
        top.wm_iconbitmap("@" + app.dataloader.findFile("pysol.xbm"))
        top.wm_iconmask("@" + app.dataloader.findFile("pysol.xbm"))
    except: pass
    app.top = top
    app.top_cursor = top.cget("cursor")

    # set global color scheme
    if os.name == "posix":              # Unix/X11
        ##color = "grey"
        color = "#d9d9d9"               # Tk 8.0p2 default
        ##top.tk_setPalette(color)
        top.tk_setPalette("background", color, "activeBackground", color)
    if os.name == "mac":
        color, priority = "#d9d9d9", "60"
        classes = (
            "Button", "Canvas", "Checkbutton", "Entry",
            "Frame", "Label", "Listbox", "Menubutton", ### "Menu",
            "Message", "Radiobutton", "Scale", "Scrollbar", "Text",
        )
        for c in classes:
            top.option_add("*" + c + "*background", color, priority)
            top.option_add("*" + c + "*activeBackground", color, priority)

    # check games
    if len(app.gdb.getGamesIdSortedByName()) == 0:
        wm_withdraw(top)
        top.update()
        d = MfxDialog(top, title=PACKAGE + " installation error",
                      text="No games were found !!!\n\nMain data directory is:\n" +
                           app.dataloader.dir + "\n\nPlease check your PySol installation.",
                      bitmap="error", strings=("Quit",))
        return 1

    # init cardsets
    app.initCardsets()
    cardset = None
    n, bn = app.opt.cardset.get(0, (None, None))
    if n:
        cardset = app.cardset_manager.getByName(n)
        if cardset and bn:
            cardset.updateCardback(backname=bn)
    if not cardset:
        cardset = app.cardset_manager.get(0)
    if app.cardset_manager.len() == 0 or not cardset:
        wm_withdraw(top)
        top.update()
        d = MfxDialog(top, title=PACKAGE + " installation error",
                      text="No cardsets were found !!!\n\nMain data directory is:\n" +
                           app.dataloader.dir + "\n\nPlease check your PySol installation.",
                      bitmap="error", strings=("Quit",))
        ##raise Exception, "no cardsets found !"
        return 1
    app.images = MfxImages(app.dataloader, None)

    # init tiles
    manager = app.tabletile_manager
    tile = Tile()
    tile.name = "None"
    tile.filename = None
    manager.register(tile)
    app.initTiles()
    if app.opt.tabletile_name and top.winfo_screendepth() > 8:
        for tile in manager.getAll():
            if app.opt.tabletile_name == tile.basename:
                app.tabletile_index = tile.index
                break
    manager.setSelected(app.tabletile_index)

    # init samples and music resources
    app.initSamples()
    app.initMusic()

    # init audio 2)
    app.audio.connectServer(app)
    if app.audio.audiodev is None:
        app.opt.sound = 0
    app.audio.updateSettings()
    # start up the background music
    if app.audio.audiodev:
        music = app.music_manager.getAll()
        if music:
            app.music_playlist = music[:]
            app.miscrandom.shuffle(app.music_playlist)
            if 1:
                for m in app.music_playlist:
                    if string.lower(m.name) == "bye_for_now":
                        app.music_playlist.remove(m)
                        app.music_playlist.insert(0, m)
                        break
            app.audio.playContinuousMusic(app.music_playlist)

    # misc. maintainter util (mainly for updating my webpages)
#%ifndef BUNDLE
    if 1 and app.debug >= 3:
        print "Games:", len(app.gdb.getGamesIdSortedByName())
        for id in app.gdb.getGamesIdSortedByName():
            print "    \"%s\"," % app.getGameMenuitemName(id)
        m = app.cardset_manager
        print "Cardsets:", m.len()
        for obj in m.getAllSortedByName():
            print "    \"%s\"," % obj.name
        m = app.tabletile_manager
        print "Table tiles:", m.len()
        for obj in m.getAllSortedByName():
            print "    \"%s\"," % obj.name
        m = app.music_manager
        print "Music songs:", m.len()
        for obj in m.getAllSortedByName():
            print "    \"%s\"," % obj.name
        sys.exit(0)
    if 1 and app.debug:
        for id in app.gdb.getGamesIdSortedByName():
            f = os.path.join(app.dataloader.dir, "html", "rules", app.getGameRulesFilename(id))
            if not os.path.isfile(f):
                print "Warning:", f, "not found"
#%endif

    # prepare the progress bar
    app.progress_bg = "#c0c0c0"
    ##app.progress_bg = None
    dir = os.path.join("images", "jokers")
    app.jokers.append(app.images.loadImage(app.dataloader.findFile("joker07_40_774.gif", dir)))
    app.jokers.append(app.images.loadImage(app.dataloader.findFile("joker08_40_774.gif", dir)))
    app.jokers.append(app.dataloader.findFile("joker07_50_774.gif", dir))
    app.jokers.append(app.dataloader.findFile("joker08_50_774.gif", dir))
    app.jokers.append(app.dataloader.findFile("joker11_100_774.gif", dir))
    app.jokers.append(app.dataloader.findFile("joker10_100.gif", dir))
    if not app.progress_images:
        app.progress_images = (app.jokers[0], app.jokers[1])
    wm_withdraw(top)

    # warn about audio problems
    warn = 0
    if not opt_nosound and os.name == "posix" and pysolsoundserver is None:
        if 1 and app.opt.sound and re.search(r"linux", sys.platform, re.I):
            warn = 1
        print PACKAGE + ": pysolsoundserver module not found, sound disabled."
        sys.stdout.flush()
    if warn:
        top.update()
        d = MfxDialog(top, title=PACKAGE + " installation problem",
                      text="The pysolsoundserver module was not found.\n\nSounds and background music will be disabled.",
                      bitmap="warning", strings=("OK",))

    # create the progress bar
    title = "Welcome to " + PACKAGE
    color = app.opt.tablecolor
    if app.tabletile_index > 0:
        color = "#008200"
    app.intro.progress = PysolProgressBar(app, top, title=title, color=color, bg=app.progress_bg, images=app.progress_images)

    # prepare other images
    dir = os.path.join("images")
    ##for f in ("noredeal", "redeal"):
    for f in ("stopsign", "redeal"):
        app.redeal_images.append(app.dataloader.findFile(f+".gif", dir))
    dir = os.path.join("images", "demo")
    for f in ("demo01", "demo02", "demo03", "demo04", "demo05"):
        app.demo_images.append(app.dataloader.findFile(f+".gif", dir))

    # load some images
    for i in range(len(app.jokers)):
        if type(app.jokers[i]) is types.StringType:
            app.jokers[i] = app.images.loadImage(app.jokers[i])
            app.intro.progress.update(step=1)
    app.jokers = tuple(app.jokers)
    for i in range(len(app.redeal_images)):
        if type(app.redeal_images[i]) is types.StringType:
            app.redeal_images[i] = app.images.loadImage(app.redeal_images[i])
            app.intro.progress.update(step=1)
    app.redeal_images = tuple(app.redeal_images)
    for i in range(len(app.demo_images)):
        if type(app.demo_images[i]) is types.StringType:
            app.demo_images[i] = app.images.loadImage(app.demo_images[i])
            app.intro.progress.update(step=1)
    app.demo_images = tuple(app.demo_images)

    # load cardset
    app.loadCardset(cardset, progress=app.intro.progress)

    # let's go - enter the mainloop
    try:
        app.mainloop()
    except KeyboardInterrupt, ex:
        print "Exiting on SIGINT."
        pass
    except StandardError, ex:
        if app.game: print app.game.gameinfo.__dict__               #bundle#
        raise                                                       #bundle#
        t = str(ex.__class__) + ": " + str(ex)
        d = MfxDialog(top, title=PACKAGE + " internal error",
                      text="Internal errror. Please report this bug:\n\n"+t,
                      strings=("Quit",), bitmap="error")

    # clean up
    app.audio.destroy()         # shut down audio
    destruct(app.audio)
    app.dumpMem("final 1")                                          #bundle#
    if app.canvas is not None:
        app.canvas.destroy()
        destruct(app.canvas)
    app.toolbar.destroy()
    destruct(app.toolbar)
    destruct(app.menubar)
    app.dumpMem("final 2")                                          #bundle#
    destruct(cyclops)                                               #bundle#
    cyclops = None                                                  #bundle#
    destruct(app)
    app = None
    top.destroy()
    destruct(top)
    top = None
    return 0


# /***********************************************************************
# // main
# ************************************************************************/

def main(args=None):
    if args is None: args = ["./pysol"]                             #bundle#

    # setup (mainly for JPython)
    if not hasattr(sys, "platform"):
        sys.platform = "unknown"
    if not hasattr(sys, "executable"):
        sys.executable = None
    if not hasattr(os, "defpath"):
        os.defpath = ""

    # check versions
    if sys.platform[:4] != "java":
        if sys.version[:5] < "1.5.2":
            print "%s needs Python 1.5.2 or better (you have %s)" % (PACKAGE, sys.version)
            return 1
    assert len(tkversion) == 4
    if tkname == "tk":
        import Tkinter
        if tkversion < (8, 0, 0, 0):
            print "%s needs Tcl/Tk 8.0 or better (you have %s)" % (PACKAGE, str(tkversion))
            return 1
        # check that Tkinter bindings are also at version 1.5.2
        if not hasattr(Tkinter.Wm, "wm_aspect") or not hasattr(Tkinter.Canvas, "tag_lower"):
            print "%s: please update the Python-Tk bindings (aka Tkinter) to version 1.5.2 or better" % (PACKAGE,)
            return 1
    # check Python
    if -1 % 13 != 12:
        raise Exception

    # run it
    r = pysol_main(args)
    ##print "FINAL\n"; dumpmem()
    return r


if __name__ == "__main__":
    sys.exit(main(sys.argv))

