## vim:ts=4:et:nowrap
##
##---------------------------------------------------------------------------##
##
## PySol -- a Python Solitaire game
##
## Copyright (C) 2000 Markus Franz Xaver Johannes Oberhumer
## Copyright (C) 1999 Markus Franz Xaver Johannes Oberhumer
## Copyright (C) 1998 Markus Franz Xaver Johannes Oberhumer
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; see the file COPYING.
## If not, write to the Free Software Foundation, Inc.,
## 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
##
## Markus F.X.J. Oberhumer
## <markus.oberhumer@jk.uni-linz.ac.at>
## http://wildsau.idv.uni-linz.ac.at/mfx/pysol.html
##
##---------------------------------------------------------------------------##


# imports
import math, os, re, string, sys, types
import Tkinter, tkColorChooser, tkFileDialog

# PySol imports
from mfxutil import destruct, Struct, kwdefault
from util import PACKAGE, VERSION
from gamedb import GI
from actions import PysolMenubarActions

# toolkit imports
from tkconst import EVENT_HANDLED, EVENT_PROPAGATE
from tkutil import bind
from selectgame import SelectGameDialog, SelectGameDialogWithPreview
from selectcardset import SelectCardsetDialogWithPreview
from selectcardset import SelectCardsetByTypeDialogWithPreview
from selecttile import SelectTileDialogWithPreview
from tkstats import StatsDialog, StatsLogDialog, StatsSessionLogDialog


# /***********************************************************************
# //
# ************************************************************************/

class MfxMenubar(Tkinter.Menu):
    addPath = None

    def __init__(self, master, **kw):
        assert kw.get("name")
        self.n = kw["tearoff"] = int(kw.get("tearoff", 0))
        apply(Tkinter.Menu.__init__, (self, master, ), kw)

    def labeltoname(self, label):
        underline = -1
        m = re.search(r"^(.*)\&([^\&].*)$", label)
        if m:
            l1, l2 = m.group(1), m.group(2)
            l1 = re.sub(r"\&\&", "&", l1)
            l2 = re.sub(r"\&\&", "&", l2)
            label = l1 + l2
            underline = len(l1)
        name = string.lower(re.sub(r"[^0-9a-zA-Z]", "", label))
        return name, label, underline

    def add(self, itemType, cnf={}):
        label = cnf.get("label")
        if label:
            name, label, underline = self.labeltoname(label)
            cnf["label"] = label
            cnf["underline"] = cnf.get("underline", underline)
            path = str(self._w) + "." + name
            if self.addPath:
                self.addPath(path, self, self.n, cnf.get("menu"))
        Tkinter.Menu.add(self, itemType, cnf)
        self.n = self.n + 1


class MfxMenu(MfxMenubar):
    def __init__(self, master, label, underline=None, **kw):
        name, label, label_underline = self.labeltoname(label)
        kwdefault(kw, name=name)
        apply(MfxMenubar.__init__, (self, master,), kw)
        if underline is None:
            underline = label_underline
        master.add_cascade(menu=self, label=label, underline=underline)


# /***********************************************************************
# // - create menubar
# // - update menubar
# // - menu actions
# ************************************************************************/

class PysolMenubar(PysolMenubarActions):
    def __init__(self, app, top):
        PysolMenubarActions.__init__(self, app, top)
        # init columnbreak
        sh = self.top.winfo_screenheight()
        self.__cb_max = 25
        if sh >= 600: self.__cb_max = 30
        if sh >= 768: self.__cb_max = 35
        # create menus
        self.__menubar = None
        self.__menupath = {}
        self.__keybindings = {}
        self._createMenubar()
        # set the menubar
        self.updateBackgroundImages()
        self.top.config(menu=self.__menubar)


    # create a GTK-like path
    def _addPath(self, path, menu, index, submenu):
        if not self.__menupath.has_key(path):
            ##print path, menu, index, submenu
            self.__menupath[path] = (menu, index, submenu)

    #
    # create the menubar
    #

    def _createMenubar(self):
        MfxMenubar.addPath = self._addPath
        self.__menubar = MfxMenubar(self.top, name="menubar")

        # init keybindings
        bind(self.top, "<KeyPress>", self._keyPressHandler)

        m = "Ctrl-"
        if os.name == "mac": m = "Cmd-"

        menu = self._createMenu("&File")
        menu.add_command(label="&New game", command=self.mNewGame, accelerator="N")
        menu.add_command(label="Select &random game", command=self.mSelectRandomGame, accelerator=m+"R")
        menu.add_command(label="Select game by nu&mber...", command=self.mSelectGameById, accelerator=m+"M")
        menu.add_separator()
        menu.add_command(label="&Open...", command=self.mOpen, accelerator=m+"O")
        menu.add_command(label="&Save", command=self.mSave, accelerator=m+"S")
        menu.add_command(label="Save &as...", command=self.mSaveAs)
        menu.add_separator()
        menu.add_command(label="&Hold and quit", command=self.mHoldAndQuit)
        menu.add_command(label="&Quit", command=self.mQuit, accelerator=m+"Q")

        menu = self._createMenu("&Select")
        self._addSelectGameMenu(menu)

        menu = self._createMenu("&Edit")
        menu.add_command(label="&Undo", command=self.mUndo, accelerator="Z")
        menu.add_command(label="&Redo", command=self.mRedo, accelerator="R")
        menu.add_command(label="Redo &all", command=self.mRedoAll)
        menu.add_separator()
        menu.add_command(label="Restart &game", command=self.mRestart, accelerator=m+"G")

        menu = self._createMenu("&Game")
        menu.add_command(label="&Deal cards", command=self.mDeal, accelerator="D")
        menu.add_command(label="&Auto drop", command=self.mDrop, accelerator="A")
        menu.add_separator()
        menu.add_command(label="S&tatus...", command=self.mStatus, accelerator="T")

        menu = self._createMenu("&Assist")
        menu.add_command(label="&Hint", command=self.mHint, accelerator="H")
        menu.add_command(label="Highlight p&iles", command=self.mHighlightPiles, accelerator="I")
        menu.add_separator()
        menu.add_command(label="&Demo", command=self.mDemo, accelerator=m+"D")
        menu.add_command(label="Demo (&all games)", command=self.mMixedDemo)

        menu = self._createMenu("&Options")
##        menu.add_command(label="&Player name...", command=self.mOptPlayerName)
        menu.add_checkbutton(label="&Confirm quit", variable=self.tkopt.confirm, command=self.mOptConfirm)
        submenu = MfxMenu(menu, label="&Automatic play")
        submenu.add_checkbutton(label="Auto &face up", variable=self.tkopt.autofaceup, command=self.mOptAutoFaceUp)
        submenu.add_checkbutton(label="&Auto drop", variable=self.tkopt.autodrop, command=self.mOptAutoDrop)
        submenu.add_checkbutton(label="Auto &deal", variable=self.tkopt.autodeal, command=self.mOptAutoDeal)
        submenu.add_separator()
        submenu.add_checkbutton(label="&Quick play", variable=self.tkopt.quickplay, command=self.mOptQuickPlay)
        submenu = MfxMenu(menu, label="Assist &level")
        submenu.add_checkbutton(label="Enable &undo", variable=self.tkopt.undo, command=self.mOptEnableUndo)
        submenu.add_checkbutton(label="Enable &hint", variable=self.tkopt.hint, command=self.mOptEnableHint)
        submenu.add_checkbutton(label="Enable highlight &piles", variable=self.tkopt.highlight_piles, command=self.mOptEnableHighlightPiles)
        submenu.add_checkbutton(label="Enable highlight &cards", variable=self.tkopt.highlight_cards, command=self.mOptEnableHighlightCards)
        submenu.add_checkbutton(label="Enable highlight same &rank", variable=self.tkopt.highlight_samerank, command=self.mOptEnableHighlightSameRank)
        menu.add_separator()
        if self.app.audio.audiodev is None:
            menu.add_checkbutton(label="Sound", variable=self.tkopt.sound, command=self.mOptSound, state=Tkinter.DISABLED)
        elif self.app.audio.server:
            menu.add_checkbutton(label="Sound...", variable=self.tkopt.sound, command=self.mOptSoundDialog)
        else:
            menu.add_checkbutton(label="Sound", variable=self.tkopt.sound, command=self.mOptSound)
        # cardsets
        manager = self.app.cardset_manager
        n = manager.len()
        menu.add_command(label="Cards&et...", command=self.mSelectCardsetDialog, accelerator="E")
        # this submenu will get set by updateBackgroundImages()
        submenu = MfxMenu(menu, label="Card &background")
        # table tiles
        manager = self.app.tabletile_manager
        n = manager.len()
        if 1:
            menu.add_command(label="Table t&ile...", command=self.mSelectTileDialog)
        else:
            submenu = MfxMenu(menu, label="Table t&ile")
            state = (Tkinter.DISABLED, Tkinter.NORMAL) [n > 1]
            cb = (25, self.__cb_max) [ n > 4 * 25 ]
            for i in range(n):
                obj = manager.get(i)
                columnbreak = i > 0 and (i % cb) == 0
                underline = (i == 0) - 1
                submenu.add_radiobutton(label=obj.name, variable=self.tkopt.tabletile, value=i,
                                        command=self.mOptTableTile, columnbreak=columnbreak, state=state,
                                        underline=underline)
            menu.add_command(label="Table colo&r...", command=self.mOptTableColor)
        submenu = MfxMenu(menu, label="A&nimations")
        submenu.add_radiobutton(label="&None", variable=self.tkopt.animations, value=0, command=self.mOptAnimations)
        submenu.add_radiobutton(label="&Timer based", variable=self.tkopt.animations, value=2, command=self.mOptAnimations)
        submenu.add_radiobutton(label="&Fast", variable=self.tkopt.animations, value=1, command=self.mOptAnimations)
        submenu.add_radiobutton(label="&Slow", variable=self.tkopt.animations, value=3, command=self.mOptAnimations)
        submenu.add_radiobutton(label="&Very slow", variable=self.tkopt.animations, value=4, command=self.mOptAnimations)
        menu.add_checkbutton(label="Card shado&w", variable=self.tkopt.shadow, command=self.mOptShadow)
        menu.add_checkbutton(label="Shade &legal moves", variable=self.tkopt.shade, command=self.mOptShade)
        menu.add_separator()
        menu.add_command(label="&Hint options...", command=self.mOptHintOptions)
        menu.add_command(label="&Demo options...", command=self.mOptDemoOptions)
        menu.add_separator()
        submenu = MfxMenu(menu, label="&Toolbar")
        submenu.add_radiobutton(label="&Hide", variable=self.tkopt.toolbar, value=0, command=self.mOptToolbar)
        submenu.add_radiobutton(label="&Top", variable=self.tkopt.toolbar, value=1, command=self.mOptToolbar)
        submenu.add_radiobutton(label="&Bottom", variable=self.tkopt.toolbar, value=2, command=self.mOptToolbar)
        menu.add_checkbutton(label="Stat&usbar", variable=self.tkopt.statusbar, command=self.mOptStatusbar)
###        menu.add_separator()
###        menu.add_command(label="&Save options", command=self.mOptSave)

        menu = self._createMenu("&Help")
        menu.add_command(label="&Contents", command=self.mHelp, accelerator=m+"F1")
        menu.add_command(label="&How to play", command=self.mHelpHowToPlay)
        menu.add_command(label="&Rules for this game", command=self.mHelpRules, accelerator="F1")
        menu.add_command(label="&License terms", command=self.mHelpLicense)
        if os.name == "nt":
            menu.add_separator()
            menu.add_command(label="PySol &Web Site", command=self.mHelpWebSite)
        menu.add_separator()
        menu.add_command(label="&About PySol...", command=self.mHelpAbout)

        if os.name == "mac":
            menu = self._createMenu("Apple")
            menu.add_command(label="About PySol...", command=self.mHelpAbout)
            menu.add_command(label="PySol help", command=self.mHelp)

        MfxMenubar.addPath = None

        ### FIXME: all key bindings should be *added* to keyPressHandler
        ctrl = "Control-"
        if os.name == "mac": ctrl = "Command-"
        self._bindKey("",   "n", self.mNewGame)
        self._bindKey("",   "g", self.mSelectGameDialog)
        self._bindKey("",   "v", self.mSelectGameDialogWithPreview)
        self._bindKey(ctrl, "r", self.mSelectRandomGame)
        self._bindKey(ctrl, "m", self.mSelectGameById)
        self._bindKey(ctrl, "n", self.mNewGameWithNextId)
        self._bindKey(ctrl, "o", self.mOpen)
        ##self._bindKey("",   "F3", self.mOpen)           # undocumented
        self._bindKey(ctrl, "s", self.mSave)
        ##self._bindKey("",   "F2", self.mSaveAs)         # undocumented
        self._bindKey(ctrl, "q", self.mQuit)
        self._bindKey("",   "z", self.mUndo)
        self._bindKey("",   "BackSpace", self.mUndo)    # undocumented
        self._bindKey("",   "r", self.mRedo)
        self._bindKey(ctrl, "g", self.mRestart)
        self._bindKey("",   "space", self.mDeal)        # undocumented
        self._bindKey("",   "t", self.mStatus)
        self._bindKey(ctrl, "t", self.mPlayerStats)     # undocumented
        self._bindKey("",   "h", self.mHint)
        self._bindKey(ctrl, "h", self.mHint1)           # undocumented
        ##self._bindKey("",   "Shift_L", self.mHighlightPiles)
        ##self._bindKey("",   "Shift_R", self.mHighlightPiles)
        self._bindKey("",   "i", self.mHighlightPiles)
        self._bindKey(ctrl, "d", self.mDemo)
        self._bindKey("",   "e", self.mSelectCardsetDialog)
        self._bindKey(ctrl, "b", self.mOptChangeCardback) # undocumented
        self._bindKey(ctrl, "i", self.mOptChangeTableTile) # undocumented
        self._bindKey(ctrl, "p", self.mOptPlayerName)   # undocumented
        self._bindKey(ctrl, "F1", self.mHelp)
        self._bindKey("",   "F1", self.mHelpRules)
        self._bindKey("",   "Print", self.mScreenshot)
        self._bindKey(ctrl, "u", self.mPlayNextMusic)   # undocumented
        # ASD and LKJ
        self._bindKey("",   "a", self.mDrop)
        self._bindKey(ctrl, "a", self.mDrop1)
        self._bindKey("",   "s", self.mUndo)
        self._bindKey("",   "d", self.mDeal)
        self._bindKey("",   "l", self.mDrop)
        self._bindKey(ctrl, "l", self.mDrop1)
        self._bindKey("",   "k", self.mUndo)
        self._bindKey("",   "j", self.mDeal)

        if 1 and self.app.debug:
            self._bindKey(ctrl, "End", self.mPlayNextMusic)
            self._bindKey(ctrl, "Prior", self.mSelectPrevGameByName)
            self._bindKey(ctrl, "Next", self.mSelectNextGameByName)
            self._bindKey(ctrl, "Up", self.mSelectPrevGameById)
            self._bindKey(ctrl, "Down", self.mSelectNextGameById)

    #
    # key binding utility
    #

    def _bindKey(self, modifier, key, func):
        if 0 and not modifier and len(key) == 1:
            self.__keybindings[string.lower(key)] = func
            self.__keybindings[string.upper(key)] = func
            return
        sequence = "<" + modifier + "KeyPress-" + key + ">"
        try:
            bind(self.top, sequence, func)
            if len(key) == 1 and key != string.upper(key):
                key = string.upper(key)
                sequence = "<" + modifier + "KeyPress-" + key + ">"
                bind(self.top, sequence, func)
        except:
            raise                                                   #bundle#
            pass


    def _keyPressHandler(self, event):
        r = EVENT_PROPAGATE
        if event and self.game:
            ##print event.__dict__
            if self.game.demo:
                # stop the demo by setting self.game.demo.keypress
                if event.char:    # ignore Ctrl/Shift/etc.
                    self.game.demo.keypress = event.char
                    r = EVENT_HANDLED
            func = self.__keybindings.get(event.char)
            if func and (event.state & ~2) == 0:
                func(event)
                r = EVENT_HANDLED
        return r


    # utility
    def _createMenu(self, label, tearoff=0):
        # It took me hours to find out how to place the help-menu
        # on the right side of the menubar. Finally I discovered
        # that there seems to be some magic with the widget name...
        if os.name == "posix":
            tearoff = 1
        menu = MfxMenu(self.__menubar, label, tearoff=tearoff, underline=-1)
        return menu


    #
    # Select Game menu entry
    #

    def _addSelectGameMenu(self, menu):
        games = map(self.app.gdb.get, self.app.gdb.getGamesIdSortedByShortName())
        games = tuple(games)
        ###menu = MfxMenu(menu, label="Select &game")
        menu.add_command(label="All &games...", command=self.mSelectGameDialog, accelerator="G")
        menu.add_command(label="Playable pre&view...", command=self.mSelectGameDialogWithPreview, accelerator="V")
        menu.add_separator()
        data = ("&Popular games", lambda gi: gi.si.game_flags & GI.GT_POPULAR)
        self._addSelectGameSubMenu(menu, games, (data, ),
                                   self.mSelectGamePopular, self.tkopt.gameid_popular)
        ##menu.add_separator()
        submenu = MfxMenu(menu, label="&Special games")
        self._addSelectGameSubMenu(submenu, games, GI.SELECT_SPECIAL_GAME_BY_TYPE,
                                   self.mSelectGame, self.tkopt.gameid)
        menu.add_separator()
        self._addSelectGameSubMenu(menu, games, GI.SELECT_GAME_BY_TYPE,
                                   self.mSelectGame, self.tkopt.gameid)


    def _addSelectGameSubMenu(self, menu, games, select_data, command, variable):
        ##print select_data
        need_sep = 0
        for label, select_func in select_data:
            if label is None:
                need_sep = 1
                continue
            g = filter(select_func, games)
            if not g:
                continue
            if need_sep:
                menu.add_separator()
                need_sep = 0
            submenu = MfxMenu(menu, label=label)
            cb = (25, self.__cb_max) [ len(g) > 4 * 25 ]
            for i in range(len(g)):
                gi = g[i]
                columnbreak = i > 0 and (i % cb) == 0
                submenu.add_radiobutton(command=command, variable=variable,
                                        columnbreak=columnbreak,
                                        value=gi.id, label=gi.short_name)

    def _mSelectGameDialog(self, d):
        if d.status == 0 and d.button == 0 and d.gameid != self.game.id:
            self.tkopt.gameid.set(d.gameid)
            self.tkopt.gameid_popular.set(d.gameid)
            if 0:
                self._mSelectGame(d.gameid, random=d.random)
            else:
                # don't ask areYouSure()
                self.game.endGame()
                self.game.quitGame(d.gameid, random=d.random)
        return EVENT_HANDLED

    def mSelectGameDialog(self, *event):
        if self._cancelDrag(): return
        d = SelectGameDialog(self.top, title="Select game",
                             app=self.app, gameid=self.game.id)
        return self._mSelectGameDialog(d)

    def mSelectGameDialogWithPreview(self, *event):
        if self._cancelDrag(): return
        d = SelectGameDialogWithPreview(self.top, title="Select game",
                                        app=self.app, gameid=self.game.id)
        return self._mSelectGameDialog(d)

    #
    # menubar overrides
    #

    def updateBackgroundImages(self):
        # delete all entries
        submenu = self.__menupath[".menubar.options.cardbackground"][2]
        submenu.delete(0, "last")
        # insert new cardbacks
        mbacks = self.app.images.getCardbacks()
        cb = int(math.ceil(math.sqrt(len(mbacks))))
        for i in range(len(mbacks)):
            columnbreak = i > 0 and (i % cb) == 0
            submenu.add_radiobutton(label=mbacks[i].name, image=mbacks[i].menu_image, variable=self.tkopt.cardback, value=i,
                                    command=self.mOptCardback, columnbreak=columnbreak, indicatoron=0, hidemargin=0)


    #
    # menu updates
    #

    def setMenuState(self, state, path):
        #print state, path
        s = (Tkinter.DISABLED, Tkinter.NORMAL)[state]
        path = ".menubar." + path
        menu, index, submenu = self.__menupath[path]
        menu.entryconfig(index, state=s)

    def setToolbarState(self, state, path):
        #print state, path
        s = (Tkinter.DISABLED, Tkinter.NORMAL)[state]
        w = getattr(self.app.toolbar, path + "_button")
        w["state"] = s


    #
    # menu actions
    #

    DEFAULTEXTENSION = ".pso"
    FILETYPES = (("PySol files", "*.pso"), ("All files", "*"))

    def mOpen(self, *event):
        if self._cancelDrag(): return
        filename = self.game.filename
        if filename:
            idir, ifile = os.path.split(os.path.normpath(filename))
        else:
            idir, ifile = "", ""
        if not idir:
            idir = self.app.dn.savegames
        d = tkFileDialog.Open()
        filename = d.show(filetypes=self.FILETYPES, defaultextension=self.DEFAULTEXTENSION,
                          initialdir=idir, initialfile=ifile)
        if filename:
            filename = os.path.normpath(filename)
            ##filename = os.path.normcase(filename)
            if os.path.isfile(filename):
                self.game.loadGame(filename)

    def mSaveAs(self, *event):
        if self._cancelDrag(): return
        if not self.menustate.save_as:
            return
        filename = self.game.filename
        if not filename:
            filename = self.app.getGameSaveName(self.game.id)
            if os.name == "posix":
                filename = filename + "-" + self.game.getFullId(format=0)
            else:
                filename = filename + "-01"
            filename = filename + self.DEFAULTEXTENSION
        idir, ifile = os.path.split(os.path.normpath(filename))
        if not idir:
            idir = self.app.dn.savegames
        ##print self.game.filename, ifile
        d = tkFileDialog.SaveAs()
        filename = d.show(filetypes=self.FILETYPES, defaultextension=self.DEFAULTEXTENSION,
                          initialdir=idir, initialfile=ifile)
        if filename:
            filename = os.path.normpath(filename)
            ##filename = os.path.normcase(filename)
            self.game.saveGame(filename)
            self.updateMenus()

    def _mPlayerStats(self, player, header):
        d = StatsDialog(self.top, header, self.app, player)
        if d.status == 0:
            self._mPlayerStatsAction(player, header, d.button)

    def _mPlayerLog(self, player, mode=0):
        while 1:
            if mode == 0:
                header="Full log for " + (player or "demo")
                d = StatsLogDialog(self.top, header, self.app, player)
            else:
                header="Session log for " + (player or "demo")
                d = StatsSessionLogDialog(self.top, header, self.app, player)
            if d.status == 0 and d.button == 2:
                mode = not mode
                continue
            if d.status == 0:
                self._mPlayerLogAction(player, header, mode, d.button)
            break

    def mSelectCardsetDialog(self, *event):
        if self._cancelDrag(): return
        ##strings, default = ("OK", "Load", "Cancel"), 0
        strings, default = (None, "Load", "Cancel"), 1
        key = self.app.nextgame.cardset.index
        d = SelectCardsetDialogWithPreview(self.top, title="Select cardset",
                app=self.app, manager=self.app.cardset_manager, key=key,
                strings=strings, default=default)
        cs = self.app.cardset_manager.get(d.key)
        if cs is None or d.key == self.app.cardset.index:
            return
        if d.status == 0 and d.button in (0, 1) and d.key >= 0:
            self.app.nextgame.cardset = cs
            if d.button == 1:
                if self.changed(restart=1):
                    if not self.game.areYouSure("Restart game", "Load cardset and restart\n the current game ?"): return
                self.game.endGame(restart=1)
                self.game.quitGame(restart=1)

    def _mOptCardback(self, index):
        if self._cancelDrag(): return
        cs = self.app.cardset
        old_index = cs.backindex
        cs.updateCardback(backindex=index)
        self.app.updateCardset(self.game.id)
        if cs.backindex == old_index:
            return
        for card in self.game.cards:
            image = self.app.images.getBack(card.deck)
            card.updateCardBackground(image=image)
        self.app.canvas.update_idletasks()
        self.tkopt.cardback.set(cs.backindex)
        # update options
        self.app.opt.cardset[0] = (cs.name, cs.backname)
        self.app.opt.cardset[cs.si.type] = (cs.name, cs.backname)

    def mOptCardback(self, *event):
        self._mOptCardback(self.tkopt.cardback.get())

    def mOptChangeCardback(self, *event):
        self._mOptCardback(self.app.cardset.backindex + 1)

    def _mOptTableTile(self, i):
        if self.app.tabletile_index != i:
            tile = self.app.tabletile_manager.get(i)
            if self.game.canvas.setTile(tile.filename):
                self.game.canvas.setTextColor(tile.text_color)
                if self.app.debug: self.game.updateStatus(info=tile.basename)
                self.tkopt.tabletile.set(i)
                self.app.tabletile_index = i
                self.app.opt.tabletile_name = tile.basename
                self.app.tabletile_manager.setSelected(i)

    def mOptTableTile(self, *event):
        if self._cancelDrag(): return
        self._mOptTableTile(self.tkopt.tabletile.get())

    def mOptChangeTableTile(self, *event):
        if self._cancelDrag(): return
        n = self.app.tabletile_manager.len()
        if n >= 2:
            i = (self.tkopt.tabletile.get() + 1) % n
            self._mOptTableTile(i)

    def mSelectTileDialog(self, *event):
        if self._cancelDrag(): return
        key = self.app.tabletile_index
        if key <= 0:
            key = string.lower(self.app.opt.tablecolor)
        d = SelectTileDialogWithPreview(self.top, title="Select table background",
                app=self.app, manager=self.app.tabletile_manager, key=key)
        if d.status == 0 and d.button in (0, 1):
            if type(d.key) is types.StringType:
                self._mOptTableColor(d.key)
            elif d.key > 0 and d.key != self.app.tabletile_index:
                self._mOptTableTile(d.key)

    def _mOptTableColor(self, color):
        if self.app.opt.tablecolor != color:
            self.app.opt.tablecolor = color
            self.game.canvas.config(bg=color)
            self.top.config(bg=color)
        self._mOptTableTile(0)          # "None"
        self.game.canvas.setTextColor(None)

    def mOptTableColor(self, *event):
        if self._cancelDrag(): return
        c = tkColorChooser.askcolor(initialcolor=self.app.opt.tablecolor,
                                    title="Select table color")
        if c and c[1]:
            self._mOptTableColor(c[1])

    def mOptToolbar(self, *event):
        if self._cancelDrag(): return
        self.setToolbarSide(self.tkopt.toolbar.get())

    def mOptToolbarRelief(self, *event):
        if self._cancelDrag(): return
        self.setToolbarRelief(self.tkopt.toolbar_relief.get())

    def mOptStatusbar(self, *event):
        if self._cancelDrag(): return
        if not self.app.statusbar: return
        side = self.tkopt.statusbar.get()
        self.app.opt.statusbar = side
        if self.app.statusbar.show(side):
            self.top.update_idletasks()


    #
    # toolbar support
    #

    def setToolbarSide(self, side):
        if self._cancelDrag(): return
        self.app.opt.toolbar = side
        self.tkopt.toolbar.set(side)                    # update radiobutton
        if self.app.toolbar.show(side):
            self.top.update_idletasks()

    def setToolbarRelief(self, relief):
        if self._cancelDrag(): return
        self.app.opt.toolbar_relief = relief
        self.tkopt.toolbar_relief.set(relief)           # update radiobutton
        ### FIXME
        self.top.update_idletasks()

