## vim:ts=4:et:nowrap
##
##---------------------------------------------------------------------------##
##
## PySol -- a Python Solitaire game
##
## Copyright (C) 2000 Markus Franz Xaver Johannes Oberhumer
## Copyright (C) 1999 Markus Franz Xaver Johannes Oberhumer
## Copyright (C) 1998 Markus Franz Xaver Johannes Oberhumer
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; see the file COPYING.
## If not, write to the Free Software Foundation, Inc.,
## 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
##
## Markus F.X.J. Oberhumer
## <markus.oberhumer@jk.uni-linz.ac.at>
## http://wildsau.idv.uni-linz.ac.at/mfx/pysol.html
##
##---------------------------------------------------------------------------##


# imports
import sys, os, string, time, types

# PySol imports
from version import VERSION, VERSION_DATE
from mfxutil import Pickler, Unpickler, UnpicklingError
from mfxutil import Struct, EnvError


# /***********************************************************************
# // constants
# ************************************************************************/

PACKAGE = "PySol"                                                   #bundle#

# Suits values are 0-3. This maps to colors 0-1.
SUITS = ("Club", "Spade", "Heart", "Diamond")
COLORS = ("black", "red")

# Card ranks are 0-12.  We also define symbolic names for the picture cards.
RANKS = ("Ace", "2", "3", "4", "5", "6", "7", "8", "9", "10", "Jack", "Queen", "King")
ACE = 0
JACK = 10
QUEEN = 11
KING = 12

# Special values for Stack.cap:
ANY_SUIT = -1
ANY_COLOR = -1
ANY_RANK = -1
NO_SUIT = 999999        # no card can ever match this suit
NO_COLOR = 999999       # no card can ever match this color
NO_RANK = 999999        # no card can ever match this rank

#
NO_REDEAL = 0
UNLIMITED_REDEALS = -1
VARIABLE_REDEALS = -2


# /***********************************************************************
# // simple benchmarking
# ************************************************************************/

class Timer:
    def __init__(self, msg = ""):
        self.msg = msg
        self.clock = time.time
        if os.name == "nt":
            self.clock = time.clock
        self.start = self.clock()
    def reset(self):
        self.start = self.clock()
    def get(self):
        return self.clock() - self.start
    def __repr__(self):
        return "%-20s %6.3f seconds" % (self.msg, self.clock() - self.start)


# /***********************************************************************
# // DataLoader
# ************************************************************************/

class DataLoader:
    def __init__(self, argv0, filenames, path=[]):
        self.dir = None
        if type(filenames) is types.StringType:
            filenames = (filenames,)
        assert type(filenames) in (types.TupleType, types.ListType)
        #$ init path
        path = path[:]
        head, tail = os.path.split(argv0)
        if not head:
            head = os.curdir
        path.append(head)
        path.append(os.path.join(head, "data"))
        path.append(os.path.join(head, os.pardir, "data"))          #bundle#
        #$ you can add your extra directories here
        if os.name == "posix":
            pass
        if os.name == "nt":
            pass
        if os.name == "mac":
            pass
        #$ add standard Unix directories to search path
        if 1 and os.name == "posix":
            for v in (VERSION, ""):
                for prefix in ("@prefix@", "/usr/local", "/usr"):
                    try:
                        if os.path.isdir(prefix):
                            path.append(os.path.join(prefix,"share/pysol",v))
                            path.append(os.path.join(prefix,"lib/pysol",v))
                            path.append(os.path.join(prefix,"share/games/pysol",v))
                            path.append(os.path.join(prefix,"lib/games/pysol",v))
                            path.append(os.path.join(prefix,"games/share/pysol",v))
                            path.append(os.path.join(prefix,"games/lib/pysol",v))
                    except EnvError:
                        pass
        #$ check path for valid directories
        self.path = []
        for p in path:
            if not p: continue
            try:
                np = os.path.normpath(p)
                if np and (not np in self.path) and os.path.isdir(np):
                    self.path.append(np)
            except EnvError:
                pass
        #$ now try to find all filenames along path
        for p in self.path:
            n = 0
            for filename in filenames:
                try:
                    f = os.path.join(p, filename)
                    if os.path.isfile(f):
                        n = n + 1
                    else:
                        break
                except EnvError:
                    pass
            if n == len(filenames):
                self.dir = p
                break
        else:
            raise os.error, str(argv0) + ": DataLoader could not find " + str(filenames)
        ##print path, self.path, self.dir


    def __findFile(self, func, filename, subdirs=None):
        if subdirs is None:
            subdirs = ("",)
        elif type(subdirs) is types.StringType:
            subdirs = (subdirs,)
        for dir in subdirs:
            f = os.path.join(self.dir, dir, filename)
            f = os.path.normpath(f)
            if func(f):
                return f
        raise os.error, "DataLoader could not find " + filename + " in " + self.dir + " " + str(subdirs)

    def findFile(self, filename, subdirs=None):
        return self.__findFile(os.path.isfile, filename, subdirs)

    def findDir(self, filename, subdirs=None):
        return self.__findFile(os.path.isdir, filename, subdirs)


# /***********************************************************************
# // Images
# ************************************************************************/

class ImagesCardback:
    def __init__(self, index, name, image, menu_image=None):
        if menu_image is None: menu_image = image
        self.index = index
        self.name = name
        self.image = image
        self.menu_image = menu_image


class Images:
    def __init__(self, dataloader, cs, r=1):
        self.d = dataloader
        self.cs = cs
        if cs is None:
            return
        # copy from cardset
        self.CARDW, self.CARDH, self.CARDD = cs.CARDW/r, cs.CARDH/r, cs.CARDD/r
        self.CARD_XOFFSET = 12/r
        self.CARD_YOFFSET = cs.CARD_UP_YOFFSET
        if r > 1:
            self.CARD_YOFFSET = max(10, cs.CARD_UP_YOFFSET)/r
        self.SHADOW_XOFFSET, self.SHADOW_YOFFSET = cs.SHADOW_XOFFSET/r, cs.SHADOW_YOFFSET/r
        # other
        self._shade_index = 0
        self._card = []
        self._back = []
        self._bottom = []
        self._letter = []
        self._shadow = []
        self._shade = []

    def destruct(self):
        pass

    def loadImage(self, file):
        raise SubclassResponsibility

    def __loadCard(self, filename, check_w=1, check_h=1):
        f = os.path.join(self.cs.dir, filename)
        img = self.loadImage(file=f)
        w, h = img.width(), img.height()
        if self.CARDW < 0:
            self.CARDW, self.CARDH = w, h
        else:
            if ((check_w and w != self.CARDW) or
                (check_h and h != self.CARDH)):
                raise Exception, "Invalid size %dx%d of image %s" % (w, h, f)
        return img

    def load(self, progress=None, fast=0):
        ##fast = 1                                                    #bundle#
        ext = self.cs.ext[1:]
        pstep = 0
        if progress:
            pstep = self.cs.ncards + len(self.cs.backnames) + 4 + self.cs.nbottoms
            if not fast:
                pstep = pstep + 14 + 1
            pstep = max(0, (80.0 - progress.percent) / pstep)
        # load cards
        for n in self.cs.getFaceCardNames():
            self._card.append(self.__loadCard(n + self.cs.ext))
            self._card[-1].filename = n
            if progress: progress.update(step=pstep)
        assert len(self._card) == self.cs.ncards
        # load backgrounds and bottoms
        for name in self.cs.backnames:
            im1 = self.__loadCard(name)
            r = max(self.CARDW / 40.0, self.CARDH / 60.0, 1)
            r = int(round(r))
            if r < 2: r = 2
            im2 = im1.subsample(r)
            self._back.append(ImagesCardback(len(self._back), name, im1, im2))
        if progress: progress.update(step=1)
        for rank in range(4):
            name = "l%02d.%s" % (rank + 1, ext)
            self._letter.append(self.__loadCard(name))
            if progress: progress.update(step=pstep)
        for i in range(self.cs.nbottoms):
            name = "bottom%02d.%s" % (i + 1, ext)
            self._bottom.append(self.__loadCard(name))
            if progress: progress.update(step=pstep)
        # support shadow for max. 13 cards (memory & performance reasons)
        for i in range(14):
            if fast:
                self._shadow.append(None)
            else:
                name = "shadow%02d.%s" % (i, ext)
                self._shadow.append(self.__loadCard(name, check_w=0, check_h=0))
            if progress: progress.update(step=pstep)
        if fast:
            self._shade.append(None)
        else:
            self._shade.append(self.__loadCard("shade." + ext))
        if progress: progress.update(step=pstep)
        return 1

    def getFace(self, deck, suit, rank):
        index = suit * len(self.cs.ranks) + rank
        ##print "getFace:", suit, rank, index
        return self._card[index % self.cs.ncards]

    def getBack(self, deck, suit, rank):
        index = self.cs.backindex % len(self._back)
        return self._back[index].image

    def getTalonBottom(self):
        return self._bottom[0]

    def getReserveBottom(self):
        return self._bottom[0]

    def getSuitBottom(self, suit=-1):
        assert type(suit) is types.IntType
        if suit == -1: return self._bottom[1]   # any suit
        i = 3 + suit
        if i >= len(self._bottom):
            # Trump (for Tarot type games)
            return self._bottom[1]
        return self._bottom[i]

    def getBraidBottom(self):
        return self._bottom[2]

    def getLetter(self, rank):
        assert 0 <= rank <= 3
        return self._letter[rank]

    def getShadow(self, ncards):
        assert ncards >= 0
        if ncards >= len(self._shadow):
            ##ncards = len(self._shadow) - 1
            return None
        return self._shadow[ncards]

    def getShade(self):
        return self._shade[self._shade_index]

    def getCardbacks(self):
        return self._back


# /***********************************************************************
# //
# ************************************************************************/

class SubsampledImages(Images):
    def __init__(self, images, r=2):
        Images.__init__(self, None, images.cs, r=r)
        for im in images._card:
            self._card.append(im.subsample(r))
        for back in images._back:
            im = back.image.subsample(r)
            self._back.append(ImagesCardback(len(self._back), back.name, im, im))
        for im in images._bottom:
            self._bottom.append(im.subsample(r))
        for im in images._letter:
            self._letter.append(im.subsample(r))

    def getShadow(self, ncards):
        return None

    def getShade(self):
        return None


# /***********************************************************************
# // memory util/debugging
# ************************************************************************/

cyclops = None

